-- Odyssey Copyright (C) 2014 Chris Hart, <chart@linux.com>
-- Utilizes DaiGUI, DaiUtils Copyright (C) by daihenka

-- Based on NexusMeter by Brad Smith
-- NexusMeter Copyright (C) 2014 Bradley Smith <brad@brad-smith.co.uk>
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.

-- Main

local Odyssey = {
	battles = {},
	saved_settings = {},
	target_damage_map = {},

	Units = {},
	Abilities = {},
	CurrentCombat = {},
}

local ApolloUnit = Unit
local L

local pairs = pairs
local ipairs = ipairs
local tinsert = table.insert
local tremove = table.remove
local tsort = table.sort
local max = math.max
local min = math.min
local round = function(val) return math.floor(val + .5) end
local setmetatable = setmetatable

local Event = Event
local Events = {}
local Units = Odyssey.Units
local Abilities = Odyssey.Abilities

local StopTracking = false -- used by EndCombatAfterKill
local LastDamageAction = 0
local LastUpdate = 0
local LastTimerUpdate = 0
local InCombat = false
local NeedsUpdate = false
local Permanent = false -- manual combat start

local UnitAvailabilityQueue = {}

local FilteredAbilities = {
--	[L["Ability name"]] = true,
}
local EndCombatAfterKill = {
--  "id",
}
local NewCombatAfterKill = {
--	"id",
}

function Odyssey:GetDefaultConfig()
	self.settings = {
		windows = {},
		classColors = {},
		abilityTypeColors = {},
		defaultColor = {},
		totalColor = {},
		textColor = {},
		lock = false,
		alwaysShowPlayer = false,
		showOnlyBoss = false,
		showScrollbar = false,
		showRankNumber = true,
		showPercent = true,
		showAbsolute = true,
		showShortNumber = false,
		showFooterCurrent = true,
		showFooterOverall = true,
		showSeparatePets = false,
		showBattleTotal = true,
		showHistoryTotal = true,
		mergeAbilitiesByName = true, -- Review this, do we actually need it or is there a bug?
		bg_opacity = 0.7,
		mg_opacity = 0.4,
		bar_opacity = 0.8,
		fg_opacity = 0.9,
		combat_timeout = 3,
		hide_during_combat = false,
		show_gathering_nodes = false,
		tooltips = true,
		enabled = true,
	}

	-- Class colors taken directly from website
	-- Warrior
	self.settings.classColors[GameLib.CodeEnumClass.Warrior] = {0.8, 0.1, 0.1, 1}
	-- Engineer
	self.settings.classColors[GameLib.CodeEnumClass.Engineer] = {0.9, 0.4, 0, 1}
	-- Esper
	self.settings.classColors[GameLib.CodeEnumClass.Esper] = {0.1, 0.5, 0.7, 1}
	-- Medic
	self.settings.classColors[GameLib.CodeEnumClass.Medic] = {0.65, 0.65, 0, 1}
	-- Stalker
	self.settings.classColors[GameLib.CodeEnumClass.Stalker] = {0.5, 0.1, 0.8, 1}
	-- Spellslinger
	self.settings.classColors[GameLib.CodeEnumClass.Spellslinger] = {0.2, 0.6, 0.1, 1}

	self.settings.abilityTypeColors[GameLib.CodeEnumDamageType.Magic] = {0.5, 0.1, 0.8, 1}
	self.settings.abilityTypeColors[GameLib.CodeEnumDamageType.Tech] = {0.2, 0.6, 0.1, 1}
	self.settings.abilityTypeColors[GameLib.CodeEnumDamageType.Physical] = {0.6, 0.6, 0.6, 1}
	self.settings.abilityTypeColors[GameLib.CodeEnumDamageType.HealShields] = {0.1, 0.5, 0.7, 1}
	self.settings.abilityTypeColors[GameLib.CodeEnumDamageType.Heal] = {0, 0.8, 0, 1}
	self.settings.abilityTypeColors[GameLib.CodeEnumDamageType.Fall] = {0.6, 0.6, 0.6, 1}
	self.settings.abilityTypeColors[GameLib.CodeEnumDamageType.Suffocate] = {0.6, 0.6, 0.6, 1}

	self.settings.defaultColor = {0.4, 0.3, 0.8, 1}
	self.settings.totalColor = {0.4, 0.3, 1, 1}
	self.settings.textColor = {0.9, 0.9, 0.9, 1}
	self.settings.backgroundColor = {0.15, 0.15, 0.15, 1}

	-- Report
	self.settings.report_channel = 'say'
	self.settings.report_target = 'YourNameHere'
	self.settings.report_lines = 10
end

function Odyssey.GetDefaultWindowSettings()
	local windowDefaultSettings = {}
	windowDefaultSettings.sort = "damage"
	windowDefaultSettings.rows = 8
	windowDefaultSettings.width = 300
	windowDefaultSettings.rowHeight = 18
	windowDefaultSettings.x = (Apollo.GetDisplaySize().nWidth - windowDefaultSettings.width) / 2
	windowDefaultSettings.y = Apollo.GetDisplaySize().nHeight / 2
	return windowDefaultSettings
end

local Ability = {}
Ability.__index = Ability
function Ability:new(info, damageAction)
	local self = {}
	if info.ability:GetId() ~= nil then
	  self.detail = Abilities[info.ability:GetId() .. (info.periodic and "1" or "0")]
	  self.name = self.detail.name
	end
	self.type = info.damagetype
	self.total = 0
	self.totalCrit = 0
	self.max = 0
	self.min = 0
	self.hits = 0
	self.crits = 0
	self.swings = 0
	self.filtered = 0
	self.interrupts = 0

	if damageAction then
		self.deflects = 0
		self.absorbed = 0
		self.intercepted = 0
	end

	if info.owner then
	  self.name = self.name .. " (Pet: " .. info.caster:GetName() .. ")"
	elseif info.periodic then
	  self.name = self.name .. " (Dot)"
	end

	return setmetatable(self, Ability)
end
function Ability:clone()
	local clone = {
		detail = self.detail,
		name = self.name,
		total = self.total,
		type = self.type,
		totalCrit = self.totalCrit,
		max = self.max,
		min = self.min,
		hits = self.hits,
		crits = self.crits,
		swings = self.swings,
		filtered = self.filtered,
		interrupts = self.interrupts
	}

	if self.deflects then -- damageAction
		clone.deflects = self.deflects
		clone.absorbed = self.absorbed
		clone.intercepted = self.intercepted
	end

	return setmetatable(clone, Ability)
end
function Ability:merge(otherAbility)
	self.total = self.total + otherAbility.total
	self.totalCrit = self.totalCrit + otherAbility.totalCrit
	self.max = max(self.max, otherAbility.max)
	if self.min == 0 then
		self.min = otherAbility.min
	else
		if otherAbility.min > 0 then
			self.min = min(self.min, otherAbility.min)
		end
	end
	self.hits = self.hits + otherAbility.hits
	self.crits = self.crits + otherAbility.crits
	self.swings = self.swings + otherAbility.swings
	self.filtered = self.filtered + otherAbility.filtered
	self.interrupts = self.interrupts + otherAbility.interrupts

	if otherAbility.deflects then
		self.deflects = self.deflects + otherAbility.deflects
		self.absorbed = self.absorbed + otherAbility.absorbed
		self.intercepted = self.intercepted + otherAbility.intercepted
	end
end
function Ability:add(statType, amount, info, damageAction)
	self.total = self.total + amount

	if info.crit then
		self.crits = self.crits + 1
		self.totalCrit = self.totalCrit + amount
	elseif statType == "deflects" or statType == "interrupts" then
		self[statType] = self[statType] + 1
	else
		self.hits = self.hits + 1
	end

	self.swings = self.swings + 1

	if amount > self.max then
		self.max = amount
	end

	if amount < self.min or self.min == 0 then
		self.min = amount
	end

	if damageAction then
		self.absorbed = self.absorbed + (info.damageAbsorbed or 0)
		self.intercepted = self.intercepted + (info.damageIntercepted or 0)
	end
end
function Ability:getPreparedAbilityStatData()
	local stats = {
		{ name = L["total"], value = Odyssey.numberFormat(self.total) },
		{ name = L["max"], value = Odyssey.numberFormat(self.max) },
		{ name = L["average"], value = Odyssey.numberFormat(round(self.total / self.swings)) },
		{ name = L["average crit"], value = Odyssey.numberFormat(round(self.totalCrit / max(self.crits, 1))) },
		{ name = L["min"], value = Odyssey.numberFormat(self.min) },
		{ name = L["crit rate"], value = ("%.2f%%"):format(self.crits / self.swings * 100) },
		{ name = L["swings"], value = Odyssey.numberFormat(self.swings) },
		{ name = L["hits"], value = Odyssey.numberFormat(self.hits) },
		{ name = L["crits"], value = Odyssey.numberFormat(self.crits) },
    { name = L["filtered"], value = Odyssey.numberFormat(self.filtered) },
    { name = L["interrupts"], value = Odyssey.numberFormat(self.interrupts) }
	}

	if self.deflects then
		tinsert(stats, { name = L["deflects"], value = Odyssey.numberFormat(self.deflects) })
		tinsert(stats, { name = L["absorbed"], value = Odyssey.numberFormat(self.absorbed) })
		tinsert(stats, { name = L["intercepted"], value = Odyssey.numberFormat(self.intercepted) })
	end

	return stats
end

local AbilityDetail = {}
function AbilityDetail:new(info)
	local self = {}
	self.name = info:GetName()
	self.icon = ""
	self.type = info.type or "none"
	self.filter = false
	
	if FilteredAbilities[self.name] and not info.abilityNew then
		self.filter = true
	end

	return self
end

local Unit = {}
function Unit:new(detail)
	local self = {}
	self.name = detail:GetName()
	self.player = detail:IsACharacter()
	self.dead = detail:IsDead()
	self.id = detail:GetId()
	--self.inGroup = detail:IsInYourGroup() or detail:IsThePlayer()
	self.self = false
	self.isPet = false
	self.owner = nil

	self.calling = detail:GetClassId()
	self.hostile = (GameLib.GetPlayerUnit():GetDispositionTo(detail) ~= ApolloUnit.CodeEnumDisposition.Friendly)
		
	return self
end

local Player = {}
Player.__index = Player
function Player:new(id, reduced)
	local self = {}
	self.detail = Units[id]
	self.reduced = reduced
	self.damage = 0
	self.damageTaken = 0
	self.damageAbsorbed = 0
	self.friendlyFire = 0
	self.overkill = 0
	self.heal = 0
	self.healTaken = 0
	self.overheal = 0
	self.interrupts = 0

	if reduced then
		self.abilities = {}
		self.linkedToOwner = false
		self.pets = false
		self.interactions = false
	else
		self.linkedToOwner = false
		self.pets = {}

		self.interactions = {
			damage = {},
			damageTaken = {},
			damageAbsorbed = {},
			friendlyFire = {},
			overkill = {},
			heal = {},
			healTaken = {},
			overheal = {},
			interrupts = {}
		}
	end

	return setmetatable(self, Player)
end
function Player:addStat(interactedWith, statType, stat, amount, info, damageAction)
	local ability = self:addAbility(interactedWith, statType, stat, amount, info, damageAction)

	if not ability.filter then
		self[statType] = self[statType] + amount
	end

	if statType == "damageTaken" then
		self.damageAbsorbed = self.damageAbsorbed + (info.damageAbsorbed or 0)
	end
end
function Player:addAbility(interactedWith, statType, stat, amount, info, damageAction)
	local abilityDetail = Abilities[info.ability:GetId() .. (info.periodic and "1" or "0")]
	if self.reduced then
		local ability = self.abilities[abilityDetail]
		if not ability then
			ability = Ability:new(info, damageAction)
			self.abilities[abilityDetail] = ability
		end
		ability:add(stat, amount, info, damageAction)
	end

	if not self.reduced and self.interactions then
		local key = interactedWith and interactedWith or L["Unknown"]
		local player = self.interactions[statType][key]
		if not player then
			player = self:new(interactedWith and interactedWith.id, true)
			if not player.detail then
				player.detail = { name = key }
			end
			self.interactions[statType][key] = player
		end

		player:addStat(interactedWith, statType, stat, amount, info, damageAction)
	end

	return abilityDetail
end
function Player:getStat(sort)
	local stat = self[sort]
	for i, value in ipairs(self.pets) do
		stat = stat + value[sort]
	end
	return stat
end
function Player:getInteractions(sort)
	local data = {
		interactions = {},
		count = 0,
		max = 1,
		total = 1
	}

	local interactions = {}
	for _, interaction in pairs(self.interactions[sort]) do
		tinsert(interactions, interaction)
	end

	tsort(interactions, function (a, b)
		return a[sort] > b[sort]
	end)

	for i, interaction in ipairs(interactions) do
		local value = interaction[sort]
		local name = interaction.detail.name
		if interaction.owner then
			name = interaction.owner.name
		end

		tinsert(data.interactions, {
			name = name,
			value = value,
			ref = interaction
		})

		data.total = data.total + value
		data.count = data.count + 1
	end

	if data.count > 0 then
		data.max = max(data.interactions[1].value, 1)
	end

	return data
end
function Player:getAbility(ability, sort)
	local returnAbility
	for _, interaction in pairs(self.interactions[sort]) do
		for _, ability2 in pairs(interaction.abilities) do
			if (Odyssey.settings.mergeAbilitiesByName and (ability.name == ability2.name))
				or (not Odyssey.settings.mergeAbilitiesByName and (ability.detail == ability2.detail)) then
				if not returnAbility then
					returnAbility = ability2:clone()
				else
					returnAbility:merge(ability2)
				end
			end
		end
	end
	return returnAbility
end
function Player:getAbilities(sort)
	local abilities = {}
	for _, interaction in pairs(self.interactions[sort]) do
		for _, ability in pairs(interaction.abilities) do
			local found = false
			for i, insertedAbility in ipairs(abilities) do
				if (Odyssey.settings.mergeAbilitiesByName and (insertedAbility.name == ability.name))
					or (not Odyssey.settings.mergeAbilitiesByName and (insertedAbility.detail == ability.detail)) then
					found = true
					insertedAbility:merge(ability)
				end
			end
			if not found then
				tinsert(abilities, ability:clone())
			end
		end
	end
	
	for _, pet in pairs(self.pets) do
    for _, interaction in pairs(pet.interactions[sort]) do
      for _, ability in pairs(interaction.abilities) do
        local found = false
        for i, insertedAbility in ipairs(abilities) do
          if (Odyssey.settings.mergeAbilitiesByName and (insertedAbility.name == ability.name))
            or (not Odyssey.settings.mergeAbilitiesByName and (insertedAbility.detail == ability.detail)) then
            found = true
            insertedAbility:merge(ability)
          end
        end
        if not found then
          tinsert(abilities, ability:clone())
        end
      end
    end
  end

	tsort(abilities, function (a, b)
		return a.total > b.total
	end)

	return abilities
end
function Player:getPreparedAbilityData(sort)
	local data = {
		abilities = {},
		count = 0,
		max = 1,
		total = 0
	}

	local abilities = self:getAbilities(sort)

	for i, ability in ipairs(abilities) do
		local value = ability.total

		tinsert(data.abilities, {
			value = value,
			ref = ability
		})

		data.total = data.total + value
		data.count = data.count + 1
	end

	data.total = max(data.total, 1)
	if data.count > 0 then
		data.max = max(data.abilities[1].value, 1)
	end

	return data
end
function Player:getInteractionAbilityData()
	local data = {
		abilities = {},
		count = 0,
		max = 1,
		total = 0
	}

	local abilities = {}
	for _, ability in pairs(self.abilities) do
		local found = false
		for i, insertedAbility in ipairs(abilities) do
			if (Odyssey.settings.mergeAbilitiesByName and (insertedAbility.name == ability.name))
				or (not Odyssey.settings.mergeAbilitiesByName and (insertedAbility.detail == ability.detail)) then
				found = true
				insertedAbility:merge(ability)
			end
		end
		if not found then
			tinsert(abilities, ability:clone())
		end
	end
	
	if not self.reduced and #self.pets > 0 then
		for _, pet in pairs(self.pets) do
      		for _, ability in pairs(self.abilities) do
		        local found = false
        		for i, insertedAbility in ipairs(abilities) do
					if (Odyssey.settings.mergeAbilitiesByName and (insertedAbility.name == ability.name))
            		or (not Odyssey.settings.mergeAbilitiesByName and (insertedAbility.detail == ability.detail)) then
	            		found = true
    	        		insertedAbility:merge(ability)
        	  		end
		        end
		        if not found then
          			tinsert(abilities, ability:clone())
        		end
      		end
    	end
	end

	tsort(abilities, function (a, b)
		return a.total > b.total
	end)

	--local abilities = Player:getAbilities(sort)

	for i, ability in ipairs(abilities) do
		local value = ability.total

		tinsert(data.abilities, {
			value = value,
			ref = ability
		})

		data.total = data.total + value
		data.count = data.count + 1
	end

	data.total = max(data.total, 1)
	if data.count > 0 then
		data.max = max(data.abilities[1].value, 1)
	end

	return data
end
function Player:createFakeAbility()
	local fakeAbility = Ability:new({ability = {GetId = function() return nil end}}, true)
	local backup = fakeAbility.getPreparedAbilityStatData
	local player = self
	fakeAbility.detail = AbilityDetail:new({GetName = function() return L["Total"] end})
	fakeAbility.name = L["Total"]

	function fakeAbility:getPreparedAbilityStatData(sort)
		-- force update on retrieve

		self.total = 0
		self.totalCrit = 0
		self.max = 0
		self.min = 0
		self.hits = 0
		self.crits = 0
		self.swings = 0
		self.filtered = 0
		self.interrupts = 0

		self.deflects = 0
		self.absorbed = 0
		self.intercepted = 0

		if player.interactions then -- real player
			for _, interaction in pairs(player.interactions[sort]) do
				for _, ability in pairs(interaction.abilities) do
					self:merge(ability)
				end
			end
		else -- interacted player
			for _, ability in pairs(player.abilities) do
				self:merge(ability)
			end
		end

		return backup(self)
	end

	return fakeAbility
end
function Player:getTooltip(sort)
	local result = {"<P TextColor=\"FFFFD100\">" .. L["Top 3 Abilities:"] .. "</P>"}

	local players = {}
	for _, player in pairs(self.interactions[sort]) do
		tinsert(players, player)
	end

	tsort(players, function(a, b)
		return a[sort] > b[sort]
	end)

	local abilities = self:getAbilities(sort)

	for i = 1, 3 do
		if abilities[i] then
			tinsert(result, ("   (%d%%) %s"):format(abilities[i].total / self[sort] * 100, abilities[i].name:sub(0, 16)))
		end
	end

	tinsert(result, "<P TextColor=\"FFFFD100\">" .. L["Top 3 Interactions:"] .. "</P>")

	for i = 1, 3 do
		if players[i] then
			tinsert(result, ("   (%d%%) %s"):format(players[i][sort] / self[sort] * 100, players[i].detail.name:sub(0, 16)))
		end
	end

	tinsert(result, "<P TextColor=\"FF33FF33\">&lt;" .. L["Middle-Click for interactions"] .. "&gt;</P>")

	return table.concat(result, "\n")
end


local Combat = {}
Combat.__index = Combat
function Combat:new(overall)
	local self = {}
	self.startTime = GameLib.GetGameTime()
	self.overall = overall
	self.duration = 0
	self.previousDuration = 0
	self.players = {}
	self.hostiles = {}
	self.hasBoss = false
	return setmetatable(self, Combat)
end

function Combat:endCombat(durationIsCallTime)
	local now = GameLib.GetGameTime()
	if durationIsCallTime then
		self.duration = now - self.startTime
	else
		self.duration = now - self.startTime - (now - LastDamageAction)
	end
end

function Combat:getPreparedPlayerData(sort, showNpcs)
	local data = {
		players = {},
		count = 0,
		max = 0,
		total = 0
	}

	local players = {}
	for _, player in pairs(self.players) do
		--if (showNpcs and not player.detail.player and not player.detail.isPet)
		--	or (not showNpcs and (player.detail.player and not player.detail.isPet)) then
		if showNpcs and (not player.detail.player and not player.detail.owner)
			or not showNpcs and (player.detail.player or player.detail.owner) then
				tinsert(players, player)
		end
	end

	-- Players and pets > other units + sort by stat desc
	tsort(players, function (a, b)
		return a:getStat(sort) > b:getStat(sort)
	end)

	for i, player in ipairs(players) do
		local name = player.detail.name
		local value --= player[sort]
		local valueSansPets = player[sort]
		local valuePlusPets = player:getStat(sort)

		if player.detail.owner then
			name = ("%s (%s)"):format(player.detail.name, player.detail.owner.name)
			if not Odyssey.settings.showSeparatePets then
				name = player.detail.owner.name
			end
		end

		if not Odyssey.settings.showSeparatePets then
			value = valuePlusPets
		else
			value = valueSansPets
		end

		local found = false
		for i, owner in ipairs(data.players) do
			if owner.name == name then
				found = true
			end
		end

		tinsert(data.players, {
			name = name,
			value = value,
			ref = player
		})

		if not found then
			data.total = data.total + value
			data.count = data.count + 1
		end
	end

	if data.count > 0 then
		data.max = data.players[1].value
	end

	return data
end
function Combat:getHostile()
	local players = {}
	
	if self.overall then
	   return L["Total"]
	end
	
	for _, player in pairs(self.players) do
		tinsert(players, player)
	end

	-- other units > Players and pets + sort by damage taken desc
	tsort(players, function (a, b)
		local aCond = not (a.detail.player or a.detail.isPet)
		local bCond = not (b.detail.player or b.detail.isPet)
		if aCond and bCond then
			return a.damageTaken > b.damageTaken
		end
		return aCond and not bCond
	end)

	if #players > 0 then
		return players[1].detail.name
	end

	return ""
end

function Combat:addPlayer(id)
	local unit = Units[id]
	local player = self.players[unit]
	if not player then
		player = Player:new(id, false)
		self.players[unit] = player
	end
	if player.detail.isPet then
		self:linkPetToOwner(player)
	end

	return player
end

function Combat:linkPetToOwner(pet)
	if not pet.linkedToOwner then
		local owner = self.players[pet.detail.owner]
		if owner then
			pet.linkedToOwner = true
			tinsert(owner.pets, pet)
		end
	end
end

local function AddGlobalAbility(info)
	local key = info.ability:GetId() .. (info.periodic and "1" or "0") 
	local ability = Abilities[key]
	if not ability then
		Abilities[key] = {
		  name = info.ability:GetName(),
		  icon = info.ability:GetIcon(),
		}
	end
	return ability
end

local function AddGlobalUnit(detail, owner)
	local id = detail:GetId()
	local unit = Units[id]
	if not unit then
		unit = Unit:new(detail)
		
		-- search for pet owner if pet
		if owner then -- friendly?
			unit.isPet = true
			unit.owner = AddGlobalUnit(owner, nil)
			--unit.inGroup = unit.owner.inGroup
		end
		
		if detail:IsThePlayer() then
			unit.self = true
		end
		
		Units[id] = unit
		return unit
	end

	return unit
end

function Odyssey:GetMaxValueCombat(sort, showNpcs)
	local maxvalue = 1
	for i, combat in ipairs(self.battles) do
		local value = combat:getPreparedPlayerData(sort, showNpcs).total / max(combat.duration, 1)
		if value > maxvalue then
			maxvalue = value
		end
	end
	return maxvalue
end

local function GroupInCombat()
	if not GameLib.GetPlayerUnit() then
		-- No player
		return false
	end

	local self_InCombat = GameLib.GetPlayerUnit():IsInCombat() or Permanent
	local count = GroupLib.GetMemberCount()
	if count == 0 then
		return self_InCombat
	end

	local combat = false
	for i = 1, count do
		local unit = GroupLib.GetUnitForGroupMember(i)

		if unit and ( unit:IsInCombat() or (InCombat and unit:IsDead()) ) then
			combat = true
			break
		end
	end

	return combat
end

function Odyssey:NewCombat(permanent)
	if InCombat then
		return
	end

	self.timerPulse:Start()

	if not self.overallCombat then
		self.overallCombat = Combat:new(true)
		tinsert(self.battles, 1, self.overallCombat)
	end

	self.CurrentCombat = Combat:new(false)
	InCombat = true
	Permanent = not not permanent
	tinsert(self.battles, self.CurrentCombat)
	self.target_damage_map = {}
	self.UI.NewCombat()

	if self.settings.hide_during_combat then
		self.settings.enabled = true
		self:Off()
	end
end

function Odyssey:EndCombat(durationIsCallTime)
	if not InCombat then
		return
	end

	self.CurrentCombat:endCombat(durationIsCallTime)
	self.overallCombat.previousDuration = self.overallCombat.previousDuration + self.CurrentCombat.duration
	self.overallCombat.duration = self.overallCombat.previousDuration
	StopTracking = false
	InCombat = false
	Permanent = false
	self.target_damage_map = {}
	self.UI.Update()
	self.UI.EndCombat()

	if self.settings.hide_during_combat then
		self.settings.enabled = false
		self:On()
	end

	self.timerPulse:Stop()
	collectgarbage('collect')
end

local function CombatEventsHandler(info, statType, damageAction)
	if StopTracking and not Permanent then -- force a combat end on several bosses with incoming damage after kill extending the duration
		return
	end

  -- Sometimes the API gives us a nil target or caster, ignore these cases.
	if not info.caster or not info.target then return end

	local target = AddGlobalUnit(info.target, nil)
	local caster = AddGlobalUnit(info.caster, info.owner)

	AddGlobalAbility(info)

	local selfAction = info.caster == info.target
	if not selfAction then
		if statType == "damage" or statType == "interrupts" then
			if not InCombat then --and caster.inGroup then
				Odyssey:NewCombat()
			end
		end
	else
		-- self damage action, e.g. fall damage
		if statType == "damage" then
			statType = "damageTaken"
		end
	end

	if InCombat then
		if damageAction then-- and caster.inGroup then
			LastDamageAction = GameLib.GetGameTime()
			NeedsUpdate = true
		end
	else
		return
	end
	
	if info.caster:GetRank() == ApolloUnit.CodeEnumRank.Elite or info.target:GetRank() == ApolloUnit.CodeEnumRank.Elite then
		Odyssey.CurrentCombat.hasBoss = true
	end

	local amount = info[statType] or 0

	local overallCasterInCombat = Odyssey.overallCombat:addPlayer(info.caster:GetId())
	local casterInCombat = Odyssey.CurrentCombat:addPlayer(info.caster:GetId())

	-- Add stat to caster
	local stat
	if statType == "damageTaken" then
		stat = "damageTaken"
		amount = info.damage or 0
	elseif damageAction then
		stat = "damage"
	else
		stat = "heal"
	end
	
	overallCasterInCombat:addStat(target, stat, statType, amount, info, damageAction)
	casterInCombat:addStat(target, stat, statType, amount, info, damageAction)

	-- Add Overheal/Overkill
	local key
	if info.overkill and info.overkill > 0 then
		key = "overkill"
	elseif info.overheal and info.overheal > 0 then
		key = "overheal"
	elseif info.interrupts and info.interrupts > 0 then
	  key = "interrupts"
	end

	if key then
		overallCasterInCombat:addStat(target, key, key, info[key], info, damageAction)
		casterInCombat:addStat(target, key, key, info[key], info, damageAction)
	end

	-- Add targets equivalent damage/heal taken
	if target then
		local overallTargetInCombat = Odyssey.overallCombat:addPlayer(info.target:GetId())
		local targetInCombat = Odyssey.CurrentCombat:addPlayer(info.target:GetId())

		local taken
		if damageAction then
			taken = "damageTaken"
		elseif statType == "heal" then
			taken = "healTaken"
		end

		if taken then
			overallTargetInCombat:addStat(caster, taken, taken, amount, info, damageAction)
			targetInCombat:addStat(caster, taken, taken, amount, info, damageAction)
		end
	end

	--local dead = GameLib.GetPlayerUnit():IsDead(info.caster)
	--if dead and Odyssey.settings.stopOnDeath == true then
	--	EndCombat()
	--end

	--Print ( tostring( info.target:GetHealth() ) )
	--Print ( tostring( info.target:GetHealth() ) )

end

-- TODO: This is for tracking mob deaths for EndCombatAfterKill and NewCombatAfterKill, don't need yet
-- TODO: Add death log
--function Events.Death(info)
--	if not InCombat or StopTracking or not info.target then
--		return
--	end
--
--	local detail = InspectUnitDetail(info.target)
--	if not detail or not detail.type then
--		return
--	end
--	for i, type in ipairs(EndCombatAfterKill) do
--		if type == detail.type then
--			EndCombat()
--			StopTracking = true
--			InCombat = true
--			return
--		end
--	end
--	for i, type in ipairs(NewCombatAfterKill) do
--		if type == detail.type then
--			EndCombat()
--			return
--		end
--	end
--end

function Odyssey:On()
	if self.settings.enabled then
		return
	end

	self.settings.enabled = true
	self.UI.Visible(true)

	--self.timerPulse:Start()
	--RegisterEvent(Event.Combat.Death, 				{Events.Death,  		Info.identifier, "Handle Deaths"})
end

function Odyssey:Off()
	if not self.settings.enabled then
		return
	end
	
	self.settings.enabled = false
	self.UI.Visible(false)

	--self.timerPulse:Stop()
	--UnregisterEvent(Event.Combat.Death, 		Events.Death)
end

function Odyssey:Toggle()
	if self.settings.enabled then
		self:Off()
	else
		self:On()
	end
end

function Odyssey:Reset()
	self:EndCombat()

	Units = {}
	Abilities = {}
	self.CurrentCombat = {}
	self.overallCombat = nil

	InCombat = false
	NeedsUpdate = false

	self.battles = {}
	self.target_damage_map = {}

	self.UI.Reset()
end

local SlashCommands = setmetatable({
	show = function ()
		Odyssey:On()
	end,
	hide = function ()
		Odyssey:Off()
	end,
	default = function ()
		Odyssey.UI.Default()
	end,
	lock = function ()
	  Odyssey.settings.lock = true
	  Odyssey.UI.ShowResizer(false)
	end,
	unlock = function ()
		Odyssey.settings.lock = false
		Odyssey.UI.ShowResizer(true)
	end,
	config = function ()
		Odyssey.ConfigInit()
	end,
	toggle = function ()
		if Odyssey.settings.enabled then
			Odyssey:Off()
		else
			Odyssey:On()
		end
	end
},
{
	__index = function (t, key)
		return function ()
			Print(L["Available commands:"])
			for cmd, func in pairs(t) do
				Print("- " .. cmd)
			end
		end
	end
})

function Odyssey:OnLoad()
	self:GetDefaultConfig()

  -- TODO: yuk..
	L = Odyssey.l

  -- Set console variables to allow logging to work properly
	Apollo.SetConsoleVariable("cmbtlog.disableOtherPlayers", false)

  -- Load Sprites
	Apollo.LoadSprites("Odyssey_Sprites.xml", "OdysseySprites")
  
  -- Load Forms
	self.xmlMainDoc = XmlDoc.CreateFromFile("Odyssey.xml")
	self.xmlMainDoc:RegisterCallback("OnDocLoaded", self)
end

function Odyssey:OnDocLoaded()
	Apollo.RegisterEventHandler("WindowManagementReady",	"OnWindowManagementReady", self)
	Apollo.RegisterEventHandler("ToggleOdyssey",			"Toggle", self)
	Apollo.RegisterEventHandler("InterfaceMenuListHasLoaded", "OnInterfaceMenuListHasLoaded", self)

	Apollo.RegisterEventHandler("CombatLogHeal",			"OnCombatLogHeal", self)
	Apollo.RegisterEventHandler("CombatLogDamage",			"OnCombatLogDamage", self)
	Apollo.RegisterEventHandler("CombatLogDeflect",			"OnCombatLogDeflect", self)
	Apollo.RegisterEventHandler("CombatLogTransference",	"OnCombatLogTransference", self)
	Apollo.RegisterEventHandler("CombatLogCCState",			"OnCombatLogCCState", self)
end

function Odyssey:OnSave(eType)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then
		return
	end

	local save = {}
	save.settings = self.settings
	return save
end

function Odyssey:OnRestore(eType, save)
	if not save.settings then return end
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then
		return
	end
	self.saved_settings = save
	self:LoadSavedData( save )
end

function Odyssey:LoadSavedData( save )
	if not save.settings then return end
	self.default_settings = Odyssey.deepcopy(self.settings)
	self.settings = save.settings
	self.settings.classColors = save.settings.classColors
	self.settings.abilityTypeColors = save.settings.abilityTypeColors

	-- TODO: This needs to be a deep check
	for key, value in pairs(self.default_settings) do
		if self.settings[key] == nil then
			self.settings[key] = value
		end
	end

	for key, value in pairs(self.default_settings.classColors) do
		if self.settings.classColors[key] == nil then
			self.settings.classColors[key] = value
		end
	end

	for key, value in pairs(self.default_settings.abilityTypeColors) do
		if self.settings.abilityTypeColors[key] == nil then
			self.settings.abilityTypeColors[key] = value
		end
	end

end

function Odyssey:OnWindowManagementReady()
	-- Create update timer
	self.timerPulse = ApolloTimer.Create(0.1, true, "Update", self)
  
	-- Register slash handler
	Apollo.RegisterSlashCommand("om", "SlashHandler", self)

	-- If we have no settings, attempt to load the saved ones again
	if not self.settings and self.saved_settings then
		self:LoadSavedData(self.saved_settings)
	end

	-- If we still have no settings, then add defaults again
	if not self.settings then
		self:GetDefaultConfig()
	end

	-- If we have no window settings, then add default window
	if #self.settings.windows == 0 then
		tinsert(self.settings.windows, self.GetDefaultWindowSettings())
	end
  
	-- Create UI windows
	self.UI.Init()
  
	local loadui = self.settings.enabled
	self.settings.enabled = false
  
	-- Turn events on
	if loadui then
		self:On()
	else
		Print((L["Type /om show to reactivate %s."]):format(Apollo.GetAddonInfo("Odyssey").strName))
		self:Off()
	end
end

function Odyssey:OnInterfaceMenuListHasLoaded()
	Event_FireGenericEvent( "InterfaceMenuList_NewAddOn", "Odyssey",
		{ "ToggleOdyssey", '', 'OdysseySprites:OdysseyIcon'}
	)
end

function Odyssey:UpdateInterfaceMenuAlerts( glow, msg, num )
	if glow == nil then
		glow = false
	end
	Event_FireGenericEvent("InterfaceMenuList_AlertAddOn", "Odyssey", { glow, msg, num })
end

function Odyssey:OnCombatLogHeal(tEventArgs)
	local info = {
		target = tEventArgs.unitTarget,
		caster = tEventArgs.unitCaster,
		owner = tEventArgs.unitCasterOwner,
		ability = tEventArgs.splCallingSpell,
		damagetype = GameLib.CodeEnumDamageType.Heal,
		heal = tEventArgs.nHealAmount, --tEventArgs.nHealAmount - tEventArgs.nOverheal,
		overheal = tEventArgs.nOverheal,
		crit = tEventArgs.eCombatResult == GameLib.CodeEnumCombatResult.Critical,
	}
	
	CombatEventsHandler(info, "heal", false)
end

function Odyssey:OnCombatLogCCState(tEventArgs)
  if tEventArgs.nInterruptArmorHit > 0 then
    local info = {
      target = tEventArgs.unitTarget,
      caster = tEventArgs.unitCaster,
      owner = tEventArgs.unitCasterOwner,
      ability = tEventArgs.splCallingSpell,
      interrupts = tEventArgs.nInterruptArmorHit,
    }
    
    CombatEventsHandler(info, "interrupts", true)
  end
end

function Odyssey:OnCombatLogDeflect(tEventArgs)
	local info = {
		target = tEventArgs.unitTarget,
		caster = tEventArgs.unitCaster,
		owner = tEventArgs.unitCasterOwner,
		ability = tEventArgs.splCallingSpell,
	}

	CombatEventsHandler(info, "deflects", true)
end

function Odyssey:OnCombatLogTransference(tEventArgs)
	local info = {
	    target = tEventArgs.unitTarget,
	    caster = tEventArgs.unitCaster,
	    owner = tEventArgs.unitCasterOwner,
	    ability = tEventArgs.splCallingSpell,
	    damagetype = tEventArgs.eDamageType,
	    damage = tEventArgs.nDamageAmount - tEventArgs.nOverkill,
	    periodic = tEventArgs.bPeriodic,
	    overkill = tEventArgs.nOverkill,
	    crit = tEventArgs.eCombatResult == GameLib.CodeEnumCombatResult.Critical,
  	}

	local obj
	if info.target then
		obj = info.target
	elseif info.caster then
		if info.owner then
			obj = info.owner
		else
			obj = info.caster
		end
	end

	if obj == nil then
		return
	end

	local key = tostring(obj:GetId())

	if self.target_damage_map[key] == nil then
		if key ~= nil then
			local new_hp = obj:GetHealth()
			if new_hp ~= nil then
				if new_hp > 0 then
					if info.owner then
						self.target_damage_map[key] = new_hp + tEventArgs.nDamageAmount
					else
						self.target_damage_map[key] = new_hp
					end
				else
					if self.target_damage_map[key] == 1 then
						self.target_damage_map[key] = nil
					else
						self.target_damage_map[key] = 1
					end
				end
			elseif new_hp == nil then
				self.target_damage_map[key] = 1
				if self.settings.show_gathering_nodes == false then
					self.target_damage_map[key] = nil
					return
				end
			end
		else
			return
		end
	end

	--Print( tostring(obj:GetName()) .. " " .. Odyssey.target_damage_map[key] .. ' - ' .. tEventArgs.nDamageAmount .. " => " .. Odyssey.target_damage_map[key] - info.damage )
	self.target_damage_map[key] = self.target_damage_map[key] - info.damage

  -- Work around an issue where overkill == damage, in this case there is actually a bug in the data given to us
  -- and we didn't have purely overkill. FIXED.
	if tEventArgs.nDamageAmount == tEventArgs.nOverkill and
		(tEventArgs.nDamageAmount > 0 and tEventArgs.nOverkill > 0) then
		--Print( key .. " d:" .. tEventArgs.nDamageAmount .. " o:" .. tEventArgs.nOverkill )

 		info.damage = tEventArgs.nDamageAmount
 		if self.target_damage_map[key] ~= nil then
 			if self.target_damage_map[key] < 0 then
 				self.target_damage_map[key] = 0--math.abs(Odyssey.target_damage_map[key])
 			end
 			info.overkill = tEventArgs.nOverkill - self.target_damage_map[key]
 		 	self.target_damage_map[key] = nil
 		end
 		info.damage = info.damage - info.overkill
 	end

  CombatEventsHandler(info, "damage", true)
  
  local info = {
    target = tEventArgs.unitTarget,
    caster = tEventArgs.unitCaster,
    owner = tEventArgs.unitCasterOwner,
    ability = tEventArgs.splCallingSpell,
    damagetype = GameLib.CodeEnumDamageType.Heal,
    heal = tEventArgs.tHealData[1].nHealAmount,
    overheal = tEventArgs.tHealData[1].nOverheal,
    crit = tEventArgs.eCombatResult == GameLib.CodeEnumCombatResult.Critical,
  }
  
  CombatEventsHandler(info, "heal", false)
end

function Odyssey:OnCombatLogDamage(tEventArgs)
	local info = {
		target = tEventArgs.unitTarget,
		caster = tEventArgs.unitCaster,
		owner = tEventArgs.unitCasterOwner,
		ability = tEventArgs.splCallingSpell,
		damagetype = tEventArgs.eDamageType,
		damage = tEventArgs.nDamageAmount - tEventArgs.nOverkill,
		periodic = tEventArgs.bPeriodic,
		overkill = tEventArgs.nOverkill,
		crit = tEventArgs.eCombatResult == GameLib.CodeEnumCombatResult.Critical,
	}
	
	if info.target and
		(info.target:GetName() == "Target Dummy" or
		info.target:GetName() == "Formidable Target Dummy" or
    	info.target:GetName() == "Weak Target Dummy") then

		--Print( tEventArgs.nRawDamage )
		info.damage = tEventArgs.nRawDamage - tEventArgs.nOverkill
	end

	local obj
	if info.target then
		obj = info.target
	elseif info.caster then
		if info.owner then
			obj = info.owner
		else
			obj = info.caster
		end
	end

	if obj == nil then
		return
	end

	local key = tostring(obj:GetId())

	if self.target_damage_map[key] == nil then
		if key ~= nil then
			local new_hp = obj:GetHealth()
			if new_hp ~= nil then
				if new_hp > 0 then
					if info.owner then
						self.target_damage_map[key] = new_hp + tEventArgs.nDamageAmount
					else
						self.target_damage_map[key] = new_hp
					end
				else
					if self.target_damage_map[key] == 1 then
						self.target_damage_map[key] = nil
					else
						self.target_damage_map[key] = 1
					end
				end
			elseif new_hp == nil then
				self.target_damage_map[key] = 1
				if self.settings.show_gathering_nodes == false then
					self.target_damage_map[key] = nil
					return
				end
			end
		else
			return
		end
	end

	--Print( tostring(obj:GetName()) .. " " .. Odyssey.target_damage_map[key] .. ' - ' .. tEventArgs.nDamageAmount .. " => " .. Odyssey.target_damage_map[key] - info.damage )
	self.target_damage_map[key] = self.target_damage_map[key] - info.damage

  -- Work around an issue where overkill == damage, in this case there is actually a bug in the data given to us
  -- and we didn't have purely overkill. FIXED.
	if tEventArgs.nDamageAmount == tEventArgs.nOverkill and
		(tEventArgs.nDamageAmount > 0 and tEventArgs.nOverkill > 0) then
		--Print( key .. " d:" .. tEventArgs.nDamageAmount .. " o:" .. tEventArgs.nOverkill )

 		info.damage = tEventArgs.nDamageAmount
 		if self.target_damage_map[key] ~= nil then
 			if self.target_damage_map[key] < 0 then
 				self.target_damage_map[key] = 0--math.abs(Odyssey.target_damage_map[key])
 			end
 			info.overkill = tEventArgs.nOverkill - self.target_damage_map[key]
 		 	self.target_damage_map[key] = nil
 		end
 		info.damage = info.damage - info.overkill
 	end
	CombatEventsHandler(info, "damage", true)
end

function Odyssey:Update()
	local now = GameLib.GetGameTime()

	if InCombat then
		if not Permanent then
			if not GroupInCombat() and now - LastDamageAction >= self.settings.combat_timeout then
				self:EndCombat()
			end
		end

		if now - LastTimerUpdate > 1 then
			LastTimerUpdate = now
			self.UI.TimerUpdate(now - self.CurrentCombat.startTime)
		end
	end

	if not NeedsUpdate and not Permanent then return end

	if now - LastUpdate > 0.3 then
		self.overallCombat.duration = self.overallCombat.previousDuration + now - self.CurrentCombat.startTime
		self.CurrentCombat.duration = now - self.CurrentCombat.startTime
		LastUpdate = now
		NeedsUpdate = false

		self.UI.Update()
	end
end

function Odyssey:SlashHandler(cmd, arg)
	local list = {}

	for param in arg:gmatch("[^%s]+") do
		tinsert(list, param)
	end
	SlashCommands[list[1]](unpack(list, 2, #list))
end

Apollo.RegisterAddon(Odyssey)