-- UI

local O = Apollo.GetAddon("Odyssey")
local Info = Apollo.GetAddonInfo("Odyssey")
O.UI = { }

local L = O.l
local NumberFormat = O.numberFormat
local FormatSeconds = O.formatSeconds

local UI = UI

local pairs = pairs
local ipairs = ipairs
local tinsert = table.insert
local tsort = table.sort
local tremove = table.remove
local setmetatable = setmetatable
local max = math.max
local min = math.min
local floor = math.floor
local ceil = math.ceil
local round = function(val) return math.floor(val + .5) end

local InCombat = false
local RelativeTo = 0
local Dummy = function() end
local AsyncRowHoverEventHelper = 0 -- e.g. mousein, mousein, mouseout :(
local Windows = {}
O.Windows = Windows
local Modes = {}
--local Sortmodes = {}
local Sortmodes = {
	"damage",
	"heal",
	"damageTaken",
	"damageAbsorbed",
	"healTaken",
	"overheal",
	"overkill",
	"interrupts",
}

local Mode = {}
Mode.__index = Mode
function Mode:new(name)
	local self = {}
	self.name = name
	self.scrollOffset = 0
	self.index = 0
	return setmetatable(self, Mode)
end

function Mode:rightClick() end
function Mode:mouse4Click() end
function Mode:mouse5Click() end
function Mode:init() end
function Mode:update() end
function Mode:onSortmodeChange(window, newSortmode) return true end
function Mode:getReportText() end
function Mode:getPinReportText() end

local Sortmode = {}
Sortmode.__index = Sortmode
function Sortmode:new()
	local self = {}

	return setmetatable(self, Sortmode)
end
function Sortmode:isCompatibleWith(mode) return false end
function Sortmode:getData(mode) return {} end


local Dialog = {}
function Dialog:init()
	self.dialog = UI.CreateFrame("Frame", "O_dialog", Context)
	self.dialog:SetVisible(false)
	self.dialog:SetWidth(350)
	self.dialog:SetHeight(100)
	self.dialog:SetBackgroundColor(0, 0, 0, .9)
	self.dialog:SetPoint("CENTER", UIParent, "CENTER", 0, -120)

	self.dialogText = UI.CreateFrame("Text", "O_dialogText", self.dialog)
	self.dialogText:SetFontSize(14)
	self.dialogText:SetPoint("TOPCENTER", self.dialog, "TOPCENTER", 0, 15)

	self.dialogButtonYes = UI.CreateFrame("RiftButton", "O_dialogButtonYes", self.dialog)
	self.dialogButtonYes:SetText("Yes")
	self.dialogButtonYes:SetPoint("BOTTOMLEFT", self.dialog, "BOTTOMLEFT", 20, -5)

	self.dialogButtonNo = UI.CreateFrame("RiftButton", "O_dialogButtonNo", self.dialog)
	self.dialogButtonNo:SetText("No")
	self.dialogButtonNo:SetPoint("BOTTOMRIGHT", self.dialog, "BOTTOMRIGHT", -20, -5)
end
function Dialog:show(text, yesLabel, noLabel, yesCallback, noCallback)
	if not self.dialog then
		self:init()
	end
	local dialog = self.dialog

	dialog:SetVisible(true)
	self.dialogText:SetText(text)
	self.dialogButtonYes:SetText(yesLabel)
	self.dialogButtonNo:SetText(noLabel)

	function self.dialogButtonYes.Event:LeftPress()
		dialog:SetVisible(false)
		yesCallback()
	end
	function self.dialogButtonNo.Event:LeftPress()
		dialog:SetVisible(false)
		noCallback()
	end
end

O.Tooltip = { }
function O.Tooltip:init()
	self.tooltip = Apollo.LoadForm(O.xmlMainDoc, "Tooltip", nil, O)
	self.tooltipText = self.tooltip:FindChild("TooltipText") 
end

function O.Tooltip:show(text, anchor, center)
	if not self.tooltip then
		self:init()
	end
	
	if not O.settings.tooltips then return end

	self.tooltipText:SetAML(text)

	local mouse = Apollo:GetMouse()
	local screen_x, screen_y = Apollo.GetScreenSize()
	--Print( x .. " " .. y .. ", " .. mouse.x .. " " .. mouse.y )

	local delta_x = screen_x - mouse.x
	local delta_y = screen_y - mouse.y

	if delta_x <= 200 then
		-- far right
		mouse.x = mouse.x - 75
	end

	self.tooltip:SetAnchorOffsets(mouse.x, mouse.y, mouse.x + 500, mouse.y + 100)
	local nTextWidth, nTextHeight = self.tooltipText:SetHeightToContentHeight()
	self.tooltip:SetAnchorOffsets(mouse.x + 10, mouse.y - 20, mouse.x + nTextWidth + 25, mouse.y + nTextHeight - 8)
	self.tooltip:Show(true)
end

function O.Tooltip:hide()
	if not self.tooltip then
		return
	end
	self.tooltip:Show(false)
end

local function RemoveOtherWindows()
	while #Windows > 1 do
		Windows[#Windows].frames.base:SetVisible(false)
		tremove(Windows, #Windows)
		tremove(O.settings.windows, #Windows)
	end
end

local function Close(window)
	if #Windows == 1 then
		Print((L["Type /om show to reactivate %s."]):format(Info.strName))
		O.Off()
		return
	end

	for i = 1, #Windows do
		if window == Windows[i] then
			Windows[i].frames.base:Show(false)
			tremove(Windows, i)
			tremove(O.settings.windows, i)
			return
		end
	end
end

local function GetClassColor(unit)
	local calling
	if unit.owner then
		calling = unit.owner.calling -- pet has no calling -> get pet owner's calling
	else
		calling = unit.calling
	end

	if calling then
		return O.settings.classColors[calling]
	else
		return O.settings.defaultColor
	end
end

local function BuildFormat(absolute, perSecond, percent)
	local args = {}
	local format = ""
	if O.settings.showAbsolute then
		tinsert(args, absolute)
		format = format .. "%s" .. (O.settings.showPercent and " (" or ", ") .. "%s"
	else
		format = format .. "%s"
	end

	tinsert(args, O.numberShortFormat(perSecond))

	if O.settings.showPercent then
		tinsert(args, percent)
		format = format .. (O.settings.showAbsolute and ", " or " ") .. (not O.settings.showAbsolute and "(" or "") .. "%.1f%%)"
	end

	return format:format(unpack(args))
end

local Window = {}
Window.__index = Window
function Window:new(settings)
	local self = {}

	self.settings = settings
	self.resizing = false
	self.frames = {}
	self.lastData = {}
	self.history = {}
	self.maxValue = 0
	self.rowCount = 0
	self.scrollOffset = 0
	self.showNpcs = false
	self.isScrollbarPresent = false
	self.selectedMode = Modes.combat
	self.selectedSortmode = nil
	self.selectedCombat = nil
	self.selectedPlayer = nil
	self.selectedPlayerDetail = nil
	self.selectedAbility = nil
	self.selectedAbilityDetail = nil

	setmetatable(self, Window)

	self:init()

	return self
end

function Window:init()
  window = self
	
	window.frames.base = Apollo.LoadForm(O.xmlMainDoc, "OdysseyForm", nil, O)
	window.frames.header = window.frames.base:FindChild("Header")
	window.frames.header:SetOpacity(O.settings.bg_opacity)
	window.frames.header:SetBGColor(O.settings.backgroundColor)
	window.frames.headerLabel = window.frames.base:FindChild("HeaderText")
	window.frames.headerLabel:SetTextColor(O.settings.textColor)
	window.frames.headerLabel:SetOpacity(O.settings.fg_opacity)

	window.frames.buttons = { }
	
	window.frames.buttons.close = window.frames.base:FindChild("ButtonClose")
	window.frames.buttons.modes = window.frames.base:FindChild("ButtonModes")
	window.frames.buttons.clear = window.frames.base:FindChild("ButtonClear")
	window.frames.buttons.report = window.frames.base:FindChild("ButtonReport")
	window.frames.buttons.history = window.frames.base:FindChild("ButtonHistory")
	window.frames.buttons.config = window.frames.base:FindChild("ButtonConfig")
	window.frames.buttons.showEnemies = window.frames.base:FindChild("ButtonEnemies")
	window.frames.buttons.showPlayers = window.frames.base:FindChild("ButtonPlayers")
	window.frames.buttons.combatStart = window.frames.base:FindChild("ButtonStart")	
	window.frames.buttons.combatEnd = window.frames.base:FindChild("ButtonStop")
	
	window.frames.background = window.frames.base:FindChild("Background")
	window.frames.background:SetOpacity(O.settings.mg_opacity)
	window.frames.background:SetBGColor(O.settings.backgroundColor)
	window.frames.opacitybackground = window.frames.base:FindChild("OpacityBackground")
	window.frames.opacitybackground:SetOpacity(O.settings.bg_opacity)
	window.frames.opacitybackground:SetBGColor(O.settings.backgroundColor)
	window.frames.rowbackground = window.frames.base:FindChild("Row:Background")
	window.frames.rowbackground:SetOpacity(O.settings.mg_opacity)

	--window.frames.row_right = window.frames.base:FindChild("Row:RightLabel")
	--window.frames.row_right:SetOpacity(1)
	--window.frames.row_left = window.frames.base:FindChild("Row:LeftLabel")
	--window.frames.row_left:SetOpacity(1)

	window.frames.reportform = Apollo.LoadForm(O.xmlMainDoc, "ReportForm", nil, O)
	window.frames.reportform:Show(false)

	window.frames.resizeLeaf = window.frames.base:FindChild("ResizeLeaf")
	window.frames.footer = window.frames.base:FindChild("Footer")
	window.frames.footer:SetOpacity( O.settings.bg_opacity )
	window.frames.footer:SetBGColor(O.settings.backgroundColor)
	window.frames.copyField = window.frames.base:FindChild("CopyField")
	window.frames.copyBackground = window.frames.base:FindChild("CopyBackground")
	window.frames.copyBackground:SetOpacity(O.settings.bg_opacity)
	--window.frames.copyBackground:SetBGColor(O.settings.backgroundColor)

	window.frames.resizerLeft = window.frames.base:FindChild("ResizeLeft")
	window.frames.resizerRight = window.frames.base:FindChild("ResizeRight")
	
	window.frames.timerLabel = window.frames.base:FindChild("TimerLabel")
	window.frames.timerLabel:SetOpacity(O.settings.fg_opacity)
	window.frames.globalStatLabel = window.frames.base:FindChild("GlobalStatLabel")
	window.frames.globalStatLabel:SetOpacity(O.settings.fg_opacity)
	
	window.selectedMode:init(window)
	window:resize()
	window:showResizer(not O.settings.lock)

	if O.settings.showScrollbar then
		window:showScrollbar(true)
	end

end

function O:OnHeaderButtonDown(wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY)
	local window = self.Windows[1]

	if eMouseButton == 0 and not O.settings.lock then
		window.pressed = true
		local mouse = Apollo.GetMouse()
		window.mouseStartX = mouse.x
		window.mouseStartY = mouse.y
		
		local anchor = {window.frames.base:GetAnchorOffsets()}
		window.attrStartX = anchor[1]
		window.attrStartY = anchor[2]
	end
end

function O:OnHeaderButtonUp(wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY)
	--O:OnHeaderTextButtonUp( wndHandler, wndControl, eMouseButton )
	window.pressed = false
end

function O:OnHeaderTextButtonUp(wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY)
	local window = self.Windows[1]
	if eMouseButton == 1 then
		if window.selectedMode == Modes.modes then
			window.selectedMode = Modes.combat
			window:setMode("combat")
		else
			window.selectedMode = Modes.modes
			window:setMode("modes")
		end
	elseif eMouseButton == 0 then
		window.pressed = false
	end
end

function O:OnHeaderMouseMove(wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY)
	local window = self.Windows[1]
	
	if window.pressed then
		local mouse = Apollo.GetMouse()
		window.settings.x = mouse.x - window.mouseStartX + window.attrStartX
		window.settings.y = mouse.y - window.mouseStartY + window.attrStartY
		
		local anchor = {window.frames.base:GetAnchorOffsets()}
		window.frames.base:SetAnchorOffsets(window.settings.x, window.settings.y,
			window.settings.x + anchor[3] - anchor[1] ,window.settings.y + anchor[4] - anchor[2])
	end
end

function O:OnFooterButtonDown(wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY)
	local window = self.Windows[1]

	if eMouseButton == 0 and not O.settings.lock then
		window.pressed = true
		local mouse = Apollo.GetMouse()
		window.mouseStartX = mouse.x
		window.mouseStartY = mouse.y
		
		local anchor = {window.frames.base:GetAnchorOffsets()}
		window.attrStartX = anchor[1]
		window.attrStartY = anchor[2]
	end
end

function O:OnFooterButtonUp(wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY)
	local window = self.Windows[1]

	if eMouseButton == 1 then
	--	window:setMode("modes")
	elseif eMouseButton == 0 then
		window.pressed = false
	end
end

function O:OnFooterMouseMove(wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY)
	local window = self.Windows[1]
	
	if window.pressed then
		local mouse = Apollo.GetMouse()
		window.settings.x = mouse.x - window.mouseStartX + window.attrStartX
		window.settings.y = mouse.y - window.mouseStartY + window.attrStartY
		
		local anchor = {window.frames.base:GetAnchorOffsets()}
		window.frames.base:SetAnchorOffsets(window.settings.x, window.settings.y,
		window.settings.x + anchor[3] - anchor[1] ,window.settings.y + anchor[4] - anchor[2])
	end
end

function O:OnButtonClose(wndHandler, wndControl, eMouseButton)
	local window = self.Windows[1]
	Close(window)
	O.Tooltip:hide()
end

function O:OnButtonReport(wndHandler, wndControl, eMouseButton)
	local window = self.Windows[1]
  	local chan = window.frames.reportform:FindChild('ReportChannelEditBox')
	local target = window.frames.reportform:FindChild('ReportTargetEditBox')
	local lines = window.frames.reportform:FindChild('ReportLinesEditBox')
	chan:SetText(O.settings.report_channel)
	target:SetText(O.settings.report_target)
	lines:SetText(O.settings.report_lines)
	window.frames.reportform:Show(true)
end  

function O:OnReportConfirmButton(wndHandler, wndControl, fNewValue, fOldValue)
	local window = self.Windows[1]
	local lastMode = window:getLastMode()
	if not lastMode then
		lastMode = Modes.combat
	end

	local chan = window.frames.reportform:FindChild('ReportChannelEditBox'):GetText()
	local target = window.frames.reportform:FindChild('ReportTargetEditBox'):GetText()
	local lines = window.frames.reportform:FindChild('ReportLinesEditBox'):GetText()

	O.settings.report_channel = chan
	O.settings.report_target = target
	O.settings.report_lines = tonumber(lines)

	local text = lastMode:getReportText(window)
	window.frames.reportform:Show(false)
	window.report(text)
end

function O:OnReportClose(wndHandler, wndControl, eMouseButton)
	local window = self.Windows[1]
	window.frames.reportform:Show(false)
end

function O:OnButtonConfig(wndHandler, wndControl, eMouseButton)
	O.ConfigInit()
end

function O:OnButtonClear(wndHandler, wndControl, eMouseButton)
	O:Reset()
end

function O:OnButtonStart(wndHandler, wndControl, eMouseButton)
	O:NewCombat(true)
end

function O:OnButtonStop(wndHandler, wndControl, eMouseButton)
	O:EndCombat(true)
end

function O:OnButtonModes(wndHandler, wndControl, eMouseButton)
	local window = self.Windows[1]
	local sortMode = ""
	local oldSortmode = window.settings.sort
	for i = 1, #Sortmodes do
		sortMode = Sortmodes[i]
		local j = 0 --oldSortmode index
		if oldSortmode == sortMode then
			j = i + 1
			if j > #Sortmodes then
				sortMode = Sortmodes[1]
			else
				sortMode = Sortmodes[j]
			end
			break
		end
	end
	window.settings.sort = sortMode
	window:setMode(window.selectedMode)
	window:update()
end

function O:OnButtonHistory(wndHandler, wndControl, eMouseButton)
	local window = self.Windows[1]
	if window.selectedMode == Modes.battle then
		window:setMode(window:getLastMode())
	else
		window:setMode("battle")
	end
	--window:update()
end

function O:OnButtonEnemies(wndHandler, wndControl, eMouseButton)
	local window = self.Windows[1]

	window.frames.buttons.showPlayers:Show(true)
	window.frames.buttons.showEnemies:Show(false)
	window.showEnemies = true
	window:update()
end

function O:OnButtonPlayers(wndHandler, wndControl, eMouseButton)
	local window = self.Windows[1]

	window.frames.buttons.showPlayers:Show(false)
	window.frames.buttons.showEnemies:Show(true)
	window.showEnemies = false
	window:update()
end
	
function O:OnRowButtonUp(wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY)
	local window = self.Windows[1]

	for i = 1, window.settings.rows do
		local row = window.frames.rows[i]
		if row.base == wndControl then
			if eMouseButton == 0 then
				row.events.leftClick()
			elseif eMouseButton == 1 then
				row.events.rightClick()
			elseif eMouseButton == 2 then
				row.events.middleClick()
			end
		end
	end
end

function O:OnBackgroundButtonUp(wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY)
	local window = self.Windows[1]

	if eMouseButton == 1 then
		window.selectedMode:rightClick(window)
	elseif eMouseButton == 4 then
		window.selectedMode:mouse4Click(window)
	elseif eMouseButton == 5 then
		window.selectedMode:mouse5Click(window)
	end
end

function O:OnBackgroundScroll(wndHandler, wndControl, nLastRelativeMouseX, nLastRelativeMouseY, fScrollAmount, bConsumeMouseWheel)
	local window = self.Windows[1]

	if fScrollAmount > 0 then
		local val = max(window.scrollOffset - 1, 0)
		if val ~= window.scrollOffset then
			window.scrollOffset = val
			window:update()
	
			if window.isScrollbarPresent then
				window.frames.scrollbar:SetPosition(window.scrollOffset)
			end
		end
	else
		local val = min(window.scrollOffset + 1, window.rowCount - window.settings.rows)
		if val ~= window.scrollOffset and val > 0 then
			window.scrollOffset = val
			window:update()
	
			if window.isScrollbarPresent then
				self.frames.scrollbar:SetPosition(window.scrollOffset)
			end
		end
	end
	
	return true
end

-- TODO: I really want to get this working with mousemove...
function O:OnWindowSizeChanged(wndHandler, wndControl)
	local window = self.Windows[1]
	
	if window and wndControl == wndHandler and not window.resizing then
		local mouse = Apollo.GetMouse()

		-- This is a hack since SizeChanged gets called on window move
		if mouse.y < window.settings.y + (window.settings.rowHeight + 1) * #window.frames.rows then return end

		-- This is also a hack, we attach this event to a leaf window, otherwise we get far too many events
		window.frames.resizeLeaf:RemoveEventHandler("WindowSizeChanged")
		
		local anchor = {window.frames.base:GetAnchorOffsets()}
		local rows = round((mouse.y - anchor[2] - 45) / (window.settings.rowHeight+1))
		if rows <= 0 then rows = 1 end
		window:setRows(rows)
		window.frames.base:SetAnchorOffsets(anchor[1], anchor[2], anchor[3], 45 + anchor[2] + (rows * (window.settings.rowHeight + 1)))
		window.settings.width = anchor[3] - anchor[1]

    	for i = 1, rows do
      		local anchor = {window.frames.rows[i].base:GetAnchorOffsets()}
      		window.frames.rows[i].base:SetAnchorOffsets(anchor[1], anchor[2], anchor[1] + window.settings.width - 2, anchor[4])
    	end
    
    	window:update()
		window.frames.resizeLeaf:AddEventHandler("WindowSizeChanged", "OnWindowSizeChanged", O)
	end
end	

-- This is yet another hack to work around bugs caused by our SizeChanged on move hack...
function O:OnResizerButtonDown(wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY, bDoubleClick, bStopPropagation)
	local window = self.Windows[1]

	local mouse = Apollo.GetMouse()
	local anchor = {window.frames.base:GetAnchorOffsets()}
	local rows = round((mouse.y - anchor[2] - 45) / (window.settings.rowHeight+1))
	if rows <= 0 then rows = 1 end
	if math.abs(rows - #window.frames.rows) > 1 then
		window:setRows(rows)
		window.frames.base:SetAnchorOffsets(anchor[1], anchor[2], anchor[3], 45 + anchor[2] + (rows * (window.settings.rowHeight + 1)))
	end
end

function O:OnHeaderMouseEnter(wndHandler, wndControl, x, y)
	local window = self.Windows[1]
	
	if wndHandler == wndControl then
	  window.frames.headerLabel:SetTextColor(ApolloColor.new(0.8, 0.8, 0.8, 1))
	end
end

function O:OnHeaderMouseExit(wndHandler, wndControl, x, y)
	local window = self.Windows[1]

  if wndHandler == wndControl then
	  window.frames.headerLabel:SetTextColor(ApolloColor.new(window.settings.textColor))
	end
end

function O:OnButtonMouseEnter(wndHandler, wndControl, x, y)
  local window = self.Windows[1]
  
  if wndControl == window.frames.buttons.close then
    O.Tooltip:show(L["Close"])
  elseif wndControl == window.frames.buttons.modes then
    O.Tooltip:show(L["Modes"])
  elseif wndControl == window.frames.buttons.clear then
    O.Tooltip:show(L["Clear data"])
  elseif wndControl == window.frames.buttons.history then
    O.Tooltip:show(L["Battle History"])
  elseif wndControl == window.frames.buttons.report then
    O.Tooltip:show(L["Report"])
  elseif wndControl == window.frames.buttons.config then
    O.Tooltip:show(L["Config"])
  elseif wndControl == window.frames.buttons.showEnemies then
    O.Tooltip:show(L["Show Enemies"])
  elseif wndControl == window.frames.buttons.showPlayers then
    O.Tooltip:show(L["Show Players"])
  elseif wndControl == window.frames.buttons.combatStart then 
    O.Tooltip:show(L["Manual Start Combat"])
  elseif wndControl == window.frames.buttons.combatEnd then
    O.Tooltip:show(L["Manual Start Combat"])
  end
end

function O:OnButtonMouseExit(wndHandler, wndControl, x, y)
  local window = self.Windows[1]

  if wndHandler == wndControl then
    O.Tooltip:hide()
  end
end

function O:OnRowMouseEnter(wndHandler, wndControl, x, y)
	local window = self.Windows[1]
	
	if wndHandler == wndControl then
	    for i = 1, window.settings.rows do
      		local row = window.frames.rows[i]
      		if row.base == wndControl then
        		row.background:SetOpacity(0.9)
        		if row.tooltip then
			    	O.Tooltip:show(row.tooltip())
		    	end
		  	end
		end
	end
end

function O:OnRowMouseExit(wndHandler, wndControl, x, y)
	local window = self.Windows[1]
	
	if wndHandler == wndControl then
		for i = 1, window.settings.rows do
	      	local row = window.frames.rows[i]
      		if row.base == wndControl then
			    row.background:SetOpacity(O.settings.bar_opacity)
	   		end
		end
		
		O.Tooltip:hide()
	end
end

function Window:clearRows(count) -- from count to rowCount will be hidden
	self.scrollOffset = 0

	local rows = self.frames.rows
	if not rows or count == 1 then return end
	if not count then
		count = #rows
	end

	for i = count, 1, -1 do
		local row = rows[i]
		row.icon:Show(false)
		row.leftLabel:SetText("")
		row.rightLabel:SetText("")
		row.base:Show(false)
	end
end
function Window:update(useOldData)
	if not useOldData then
		self.lastData, self.rowCount, self.maxValue = self.selectedMode:update(self)
	end

	self.lastData = self.lastData or {}
	self.rowCount = self.rowCount or 0
	self.maxValue = max(self.maxValue or 1, 1)

	if self.isScrollbarPresent then
		local scrollbar = self.frames.scrollbar
		local val = max(self.rowCount - self.settings.rows, 0)
		if val == 0 then
			scrollbar:SetEnabled(false)
		else
			scrollbar:SetEnabled(true)
			scrollbar:SetRange(0, val)
			scrollbar:SetPosition(self.scrollOffset) -- e.g. combat mode overrides scrollOffset
		end
	end

	local anchors = {self.frames.rows[1].base:GetAnchorOffsets()}
	local maxRowWidth = anchors[3] - anchors[1]
	for i = 1, self.settings.rows do
		local row = self.frames.rows[i]
		local data = self.lastData[i]
		local rightClick
		if data then
			-- Default values
			if not data.value then
				data.value = 0
			end
			if not data.color then
				data.color = O.settings.defaultColor
			end
			if not data.leftLabel then
				data.leftLabel = ""
			end
			if not data.rightLabel then
				data.rightLabel = ""
			end
			if not data.leftClick then
				data.leftClick = Dummy
			end
			if not data.middleClick then
				data.middleClick = Dummy
			end
			if not data.rightClick then
				rightClick = function() end --O:OnBackgroundButtonUp(self, self, 1, 0, 0) end
			else
				rightClick = function() data.rightClick(self) end
			end
			if data.icon and data.icon ~= "" then
				row.icon:SetSprite(data.icon)
				row.icon:Show(true)
			else
				row.icon:Show(false)
			end

			--Print( ApolloColor.new(data.color[1], data.color[2], data.color[3], 1) )

			row.leftLabel:SetText(tostring(data.leftLabel))
			row.leftLabel:SetTextColor(O.settings.textColor)
			row.leftLabel:SetOpacity(O.settings.fg_opacity)
			row.rightLabel:SetText(tostring(data.rightLabel))
			row.rightLabel:SetTextColor(O.settings.textColor)
			row.rightLabel:SetOpacity(O.settings.fg_opacity)
			local anchor = {row.background:GetAnchorOffsets()}
			row.background:SetAnchorOffsets(anchor[1], anchor[2], anchor[1] + max(min(maxRowWidth * data.value / self.maxValue, maxRowWidth), 0), anchor[4])
			--row.background:SetBGColor(ApolloColor.new(data.color[1], data.color[2], data.color[3], 1))
			if type(data.color) ~= 'userdata' then
				data.color = ApolloColor.new( data.color[1], data.color[2], data.color[3], 1 )
			else
				data.color = ApolloColor.new( data.color )
			end
			row.background:SetBGColor(data.color)
			row.events.leftClick = data.leftClick
			row.events.middleClick = data.middleClick
			row.events.rightClick = rightClick
  			row.tooltip = data.tooltip
			row.base:Show(true)
		else
			row.base:Show(false)
		end
	end
end
function Window.report(text)
	if not text then return end

	local chan = O.settings.report_channel
	local target = O.settings.report_target

	for i = 1, #text do
		if ( chan == 'whisper' or chan == 'tell' or chan == 'w' or chan == 't' ) then
			ChatSystemLib.Command("/" .. chan .. " " .. target .. " " .. text[i])
		else
			ChatSystemLib.Command("/" .. chan .. " " .. text[i])
		end
	end
end
function Window:setRows(count)
	local forCount = count
	local rowCount = 0
	local rowpos = 1

	if self.frames.rows then
		rowCount = #self.frames.rows
		rowpos = 1 + rowCount * (self.settings.rowHeight + 1)
	end

	if rowCount == count then
		return
	end

	count = max(count, 1)
	self.settings.rows = count

	if self.frames.rows then
		local diff = count - rowCount

		if diff < 0 then -- remove row(s)
			self:clearRows(count)
			for i = count + 1, rowCount do
 				self.frames.obscured_rows[rowCount - i + count + 1] = tremove(self.frames.rows)
			end
			O.data5 = self.frames.obscured_rows
		elseif diff > 0 then -- add row(s)
			self.scrollOffset = max(min(self.scrollOffset, self.rowCount - self.settings.rows), 0)
		elseif diff == 0 then
			forCount = 0
		end
	else
		self.frames.rows = {}
		self.frames.obscured_rows = {}
	end

	for i = rowCount + 1, forCount do
		self.frames.rows[i] = self.frames.obscured_rows[i]

		if not self.frames.rows[i] then
			self.frames.rows[i] = { }
			self.frames.rows[i].events = { }
			self.frames.rows[i].base = Apollo.LoadForm(O.xmlMainDoc, "Row", self.frames.background, O)
			self.frames.rows[i].background = self.frames.rows[i].base:FindChild("Background")
			self.frames.rows[i].icon = self.frames.rows[i].base:FindChild("Icon")
			self.frames.rows[i].rightLabel = self.frames.rows[i].base:FindChild("RightLabel")
			self.frames.rows[i].leftLabel = self.frames.rows[i].base:FindChild("LeftLabel")
			--self.frames.rows[i].opacityackground:SetOpacity(O.settings.opacity * 2)
		end

		--self.frames.rows[i].leftLabel:SetOpacity(1)
		--self.frames.rows[i].rightLabel:SetOpacity(1)

		local anchor = {self.frames.base:GetAnchorOffsets()}
		self.frames.rows[i].base:SetAnchorOffsets(1, rowpos, anchor[3] - anchor[1] - 1, rowpos + self.settings.rowHeight)
		rowpos = rowpos + self.settings.rowHeight + 1
		self.frames.rows[i].base:Show(false)
	end
	
	self.resizing = true
	local anchor = {self.frames.base:GetAnchorOffsets()}
	self.frames.base:SetAnchorOffsets(anchor[1], anchor[2], anchor[3], 45 + anchor[2] + (count * (self.settings.rowHeight + 1)))
	self.resizing = false

	self:update()
end

function Window:timerUpdate(duration, text)
	if text then
		self.frames.timerLabel:SetText( text .. " " .. FormatSeconds(duration))
		self.frames.timerLabel:SetTextColor(O.settings.textColor)
	else
		self.frames.timerLabel:SetText(FormatSeconds(duration))
		self.frames.timerLabel:SetTextColor(O.settings.textColor)
	end
end

function Window:setMode(mode, ...)
	--	RelativeTo = 0
	O.Tooltip:hide()

	local newMode = {}
	if type(mode) == "string" then
		newMode = Modes[mode]
	else
		newMode = mode
	end

	if newMode ~= self.selectedMode then -- prevent endless loop
		tinsert(self.history, self.selectedMode)
		self.selectedMode = newMode
	end

	self:clearRows()
	self.selectedMode:init(self, ...)
	self:update()

	if self.isScrollbarPresent and self.frames.scrollbar:GetEnabled() then
		self.frames.scrollbar:SetPosition(self.scrollOffset)
	end
end

function Window:getLastMode()
	-- sortmode selection is skipped (Modes.modes)
	-- two same modes can't be behind one another

	local index = #self.history
	if self.history[index] == Modes.modes then
		index = index - 1
	end
	if self.history[index] == self.selectedMode then
		index = index - 1
	end

	return self.history[max(index, 1)]
end

function Window:setTitle(title)
	self.frames.headerLabel:SetText(tostring(title))
	self.frames.headerLabel:SetTextColor(O.settings.textColor)
end
function Window:setGlobalLabel(total, dps)
	self.frames.globalStatLabel:SetText(NumberFormat(total) .. ", " .. NumberFormat(dps))
	self.frames.globalStatLabel:SetTextColor(O.settings.textColor)
end
function Window:setWidth(width)
	self.frames.base:SetWidth(width)
	self:update(true)
end

function Window:resize()
	-- TODO: /om default doesn't work
	self.resizing = true
	self.frames.base:SetAnchorOffsets(self.settings.x, self.settings.y, self.settings.x + self.settings.width, self.settings.y + 45 + (self.settings.rows * (self.settings.rowHeight + 1)))
	self.resizing = false
	
	self:setRows(self.settings.rows)
end
function Window:showResizer(state)
	self.frames.resizerLeft:Show(state)
	self.frames.resizerRight:Show(state)
	self.frames.base:SetStyle("Sizable", state)
end

-- Remove or port this
function Window:showScrollbar(state)
	local window = self
	if not self.frames.scrollbar and state then
		self.frames.scrollbar = UI.CreateFrame("RiftScrollbar", "O_scrollbar", self.frames.base)
		self.frames.scrollbar:SetPoint("TOPRIGHT", self.frames.header, "BOTTOMRIGHT")
		self.frames.scrollbar:SetPoint("BOTTOM", self.frames.footer, "TOP")
		self.frames.scrollbar:SetOrientation("vertical")
		self.frames.scrollbar:SetLayer(2)
		self.frames.scrollbar:SetThickness(4)

		self.frames.scrollbar.Event.ScrollbarChange = function (self)
			local val = round(window.frames.scrollbar:GetPosition())
			if val ~= window.scrollOffset then
				window.scrollOffset = val
				window:update()
			end
		end
	end

	if state then
		self.frames.scrollbar:SetVisible(true)
		if self.rowCount - self.settings.rows > 0 then
			self.frames.scrollbar:SetRange(0, self.rowCount - self.settings.rows)
			self.frames.scrollbar:SetPosition(self.scrollOffset)
		end
	else
		self.frames.scrollbar:SetVisible(false)
	end

	for i, row in ipairs(self.frames.rows) do
		if O.settings.showScrollbar then
			row.base:SetPoint("RIGHT", self.frames.scrollbar, "LEFT")
		else
			row.base:SetPoint("RIGHT", self.frames.background, "RIGHT", -1, nil)
		end
	end

	self.isScrollbarPresent = state

	self:update(true)
end

--
--	MODES
--

Modes.modes = Mode:new("modes")
function Modes.modes:init(window)
	window:setTitle("Odyssey: " .. L["Sort Modes"])
end

function Modes.modes:update(window)
	local rows = {}
	local limit = min(#Sortmodes, window.settings.rows)

	for i = 1, limit do
		local data = {}
		local sortMode = Sortmodes[i + window.scrollOffset]
		local lastMode = window:getLastMode()
		
		data.leftLabel = L[sortMode]
		data.rightlabel = ""
		data.value = window.settings.width - 2
		data.leftClick = function()
			window.settings.sort = Sortmodes[i + window.scrollOffset]
			--if lastMode ~= nil then
			--	window:setMode(lastMode)
			--else
				window:setMode("combat")
			--end
		end

		rows[i] = data
	end
	return rows, #Sortmodes, window.settings.width - 2
end

-- Latest Battle
Modes.combat = Mode:new("combat")
function Modes.combat:init(window, combat)
    if combat then
		window.selectedCombat = combat
    end

	self.combat_index = 1

    if not window.selectedCombat then
		if #O.battles > 0 then
			window.selectedCombat = O.battles[#O.battles]
			self.combat_index = #O.battles
		end
	else
		for i = 1, #O.battles do
			self.combat_index = i
			if O.battles[i] == window.selectedCombat then
				break
			end
		end
    end

	if #O.battles > 0 and window.selectedCombat == O.battles[1] then
		window:setTitle(L["Overall"] .. ": " .. L[window.settings.sort])
		if O.settings.showFooterCurrent == true then
			local data = O.battles[#O.battles]:getPreparedPlayerData(window.settings.sort, window.showEnemies)
			local duration = max(O.battles[#O.battles].duration, 1)
			window:timerUpdate(duration, "Current: ")
			window:setGlobalLabel(data.total, data.total / duration)
		elseif O.settings.showFooterOverall == true then
			local data = O.battles[1]:getPreparedPlayerData(window.settings.sort, window.showEnemies)
			local duration = max(O.battles[1].duration, 1)
			window:timerUpdate(duration, "Overall: ")
			window:setGlobalLabel(data.total, data.total / duration)			
		end
	elseif #O.battles > 0 and window.selectedCombat ~= O.battles[#O.battles] then
		window:setTitle( "#" .. self.combat_index - 1 .. ": " .. L[window.settings.sort])
		if O.settings.showFooterOverall == true then
			local data = O.battles[1]:getPreparedPlayerData(window.settings.sort, window.showEnemies)
			local duration = max(O.battles[1].duration, 1)
			window:timerUpdate(duration, "Overall: ")
			window:setGlobalLabel(data.total, data.total / duration)
		elseif O.settings.showFooterCurrent == true then
			local data = O.battles[#O.battles]:getPreparedPlayerData(window.settings.sort, window.showEnemies)
			local duration = max(O.battles[#O.battles].duration, 1)
			window:timerUpdate(duration, "Current: ")
			window:setGlobalLabel(data.total, data.total / duration)
		end
	else
		window:setTitle(L["Current"] .. ": " .. L[window.settings.sort])
		if O.settings.showFooterOverall == true then
			if window.selectedCombat then
				local data = O.battles[1]:getPreparedPlayerData(window.settings.sort, window.showEnemies)
				local duration = max(O.battles[1].duration, 1)
				window:timerUpdate(duration, "Overall: ")
				window:setGlobalLabel(data.total, data.total / duration)
			else
				window:timerUpdate(0, "Overall: ")
				window:setGlobalLabel(0, 0)
			end
		elseif O.settings.showFooterCurrent == true then
			if window.selectedCombat then
				local data = O.battles[#O.battles]:getPreparedPlayerData(window.settings.sort, window.showEnemies)
				local duration = max(O.battles[#O.battles].duration, 1)
				window:timerUpdate(duration, "Current: ")
				window:setGlobalLabel(data.total, data.total / duration)
			else
				window:timerUpdate(0, "Current: ")
				window:setGlobalLabel(0, 0)
			end
		end
	end
end

function Modes.combat:update(window)
	if not window.selectedCombat then return end -- Update() is called on init

	--local data = window.selectedCombat:getPreparedPlayerData(window.settings.sort, window.showEnemies)
	local data = window.selectedCombat:getPreparedPlayerData(window.settings.sort, window.showEnemies)
	local duration = max(window.selectedCombat.duration, 1)

	local rows = {}
	local selfFound = false
	local limit = min(data.count, window.settings.rows)

	O:UpdateInterfaceMenuAlerts( InCombat, self:getPinReportText(window), #O.battles - 1 )

	-- Total
	if O.settings.showBattleTotal == true then
		local orow = {}
		orow.rightLabel = BuildFormat(NumberFormat(data.total), data.total / duration, 100)
		orow.value = data.total
		orow.leftClick = function() return end
		orow.leftLabel = "T: Total (" .. FormatSeconds(duration) .. ")"
		orow.color = O.settings.totalColor
		table.insert( rows, 1, orow )
	end

	for i = 1, limit do
		local row = {}
		local player = data.players[i + window.scrollOffset]

		--Print( player.name )

		if O.settings.alwaysShowPlayer then
			if player.ref.detail.self then
				selfFound = true
			end

			if i == limit and not selfFound then
				for j = i + window.scrollOffset, data.count do
					if data.players[j].ref.detail.self then
						player = data.players[j]
						i = j - window.scrollOffset
						break
					end
				end
			end
		end

		row.leftLabel = (O.settings.showRankNumber and i + window.scrollOffset .. ". " or "") .. player.name

		if RelativeTo > 0 then
			row.rightLabel = BuildFormat(NumberFormat(player.value), player.value / duration, player.value / max(data.players[RelativeTo].value, 1) * 100)
		else
			row.rightLabel = BuildFormat(NumberFormat(player.value), player.value / duration, player.value / max(data.total, 1) * 100)
		end

		row.color = GetClassColor(player.ref.detail)
		row.value = player.value
		row.tooltip = function()
			return player.ref:getTooltip(window.settings.sort)
		end
		row.leftClick = function()
			window:setMode("abilities", player.ref)
		end
		if player.ref.interactions then
			row.middleClick = function()
				window:setMode("interactions", player.ref)
			end
		end

		tinsert(rows, row)
	end

	if #O.battles > 0 and window.selectedCombat == O.battles[1] then
		window:setTitle(L["Overall"] .. ": " .. L[window.settings.sort])
		if O.settings.showFooterCurrent == true then
			local data = O.battles[#O.battles]:getPreparedPlayerData(window.settings.sort, window.showEnemies)
			local duration = max(O.battles[#O.battles].duration, 1)
			window:timerUpdate(duration, "Current: ")
			window:setGlobalLabel(data.total, data.total / duration)
		elseif O.settings.showFooterOverall == true then
			local data = O.battles[1]:getPreparedPlayerData(window.settings.sort, window.showEnemies)
			local duration = max(O.battles[1].duration, 1)
			window:timerUpdate(duration, "Overall: ")
			window:setGlobalLabel(data.total, data.total / duration)			
		end
	elseif #O.battles > 0 and window.selectedCombat ~= O.battles[#O.battles] then
		window:setTitle( "#" .. self.combat_index - 1 .. ": " .. L[window.settings.sort])
		if O.settings.showFooterOverall == true then
			local data = O.battles[1]:getPreparedPlayerData(window.settings.sort, window.showEnemies)
			local duration = max(O.battles[1].duration, 1)
			window:timerUpdate(duration, "Overall: ")
			window:setGlobalLabel(data.total, data.total / duration)
		elseif O.settings.showFooterCurrent == true then
			local data = O.battles[#O.battles]:getPreparedPlayerData(window.settings.sort, window.showEnemies)
			local duration = max(O.battles[#O.battles].duration, 1)
			window:timerUpdate(duration, "Current: ")
			window:setGlobalLabel(data.total, data.total / duration)
		end
	else
		window:setTitle(L["Current"] .. ": " .. L[window.settings.sort])
		if O.settings.showFooterOverall == true then
			if window.selectedCombat then
				local data = O.battles[1]:getPreparedPlayerData(window.settings.sort, window.showEnemies)
				local duration = max(O.battles[1].duration, 1)
				window:timerUpdate(duration, "Overall: ")
				window:setGlobalLabel(data.total, data.total / duration)
			else
				window:timerUpdate(0, "Overall: ")
				window:setGlobalLabel(0, 0)
			end
		elseif O.settings.showFooterCurrent == true then
			if window.selectedCombat then
				local data = O.battles[#O.battles]:getPreparedPlayerData(window.settings.sort, window.showEnemies)
				local duration = max(O.battles[#O.battles].duration, 1)
				window:timerUpdate(duration, "Current: ")
				window:setGlobalLabel(data.total, data.total / duration)
			else
				window:timerUpdate(0, "Current: ")
				window:setGlobalLabel(0, 0)
			end
		end
	end

	return rows, data.count, data.max
end

function Modes.combat:rightClick(window)
	if #O.battles > 0 then
		window:setMode("battle")
	end
end

function Modes.combat:getPinReportText(window)
	if not window.selectedCombat then return end

	local data = window.selectedCombat:getPreparedPlayerData(window.settings.sort, window.showEnemies)
	local duration = max(window.selectedCombat.duration, 1)
	local text = {}
	tinsert(text, ("Battle Report"))
	tinsert(text, ("Target: %s ~ (%s)"):format(window.selectedCombat:getHostile():sub(0, 16), FormatSeconds(window.selectedCombat.duration)))
	tinsert(text, ("--------------------"))
	tinsert(text, ("Total %s: %d"):format( window.settings.sort, data.total ))
	for i = 1, min( data.count, 5 ) do
		local player = data.players[i]
		if not player.ref.detail.isPet then
			tinsert(text, ("%d: %s - %d (%d) "):format( i, player.ref.detail.name:sub(0, 16), player.value, player.value / duration))
		end
	end

	return table.concat( text, "\n" )
end

function Modes.combat:getReportText(window)
	if not window.selectedCombat then return end

	local data = window.selectedCombat:getPreparedPlayerData(window.settings.sort, window.showEnemies)
	local duration = max(window.selectedCombat.duration, 1)
	local text = {}
	tinsert(text, ("Odyssey Battle Report"))
	tinsert(text, ("Target: %s ~ (%s)"):format(window.selectedCombat:getHostile():sub(0, 16), FormatSeconds(window.selectedCombat.duration)))
	tinsert(text, ("--------------------"))
	tinsert(text, ("Total %s: %d"):format( window.settings.sort, data.total ))
	for i = 1, min( data.count, O.settings.report_lines ) do
		local player = data.players[i]
		if not player.ref.detail.isPet then
			tinsert(text, ("%d: %s - %d (%d) "):format( i, player.ref.detail.name:sub(0, 16), player.value, player.value / duration))
		end
	end

	return text
end

-- Battle History
Modes.battle = Mode:new("battle")
function Modes.battle:init(window)
--	local val = max(RowCount - O.settings.rows, 0) -- RowCount is an old value. last combat count
--	RowCount = #O.batttles -- now update
--	if self.scrollOffset == val or self.scrollOffset == 0 then -- dont scroll back to the last combats if the user is already in combats selection mode and a new combat started
--		self.scrollOffset = max(RowCount - O.settings.rows, 0) -- set offset to the current combat
--	end

	window.rowCount = #O.battles
	window.scrollOffset = max(window.rowCount - window.settings.rows, 0)
	window:setTitle(L["History"] .. ": " .. L[window.settings.sort])

	if #O.battles > 0 and window.selectedCombat == O.battles[1] then
		if O.settings.showFooterCurrent == true then
			local data = O.battles[#O.battles]:getPreparedPlayerData(window.settings.sort, window.showEnemies)
			local duration = max(O.battles[#O.battles].duration, 1)
			window:timerUpdate(duration, "Current: ")
			window:setGlobalLabel(data.total, data.total / duration)
		elseif O.settings.showFooterOverall == true then
			local data = O.battles[1]:getPreparedPlayerData(window.settings.sort, window.showEnemies)
			local duration = max(O.battles[1].duration, 1)
			window:timerUpdate(duration, "Overall: ")
			window:setGlobalLabel(data.total, data.total / duration)			
		end
	elseif #O.battles > 0 and window.selectedCombat ~= O.battles[#O.battles] then
		if O.settings.showFooterOverall == true then
			local data = O.battles[1]:getPreparedPlayerData(window.settings.sort, window.showEnemies)
			local duration = max(O.battles[1].duration, 1)
			window:timerUpdate(duration, "Overall: ")
			window:setGlobalLabel(data.total, data.total / duration)
		elseif O.settings.showFooterCurrent == true then
			local data = O.battles[#O.battles]:getPreparedPlayerData(window.settings.sort, window.showEnemies)
			local duration = max(O.battles[#O.battles].duration, 1)
			window:timerUpdate(duration, "Current: ")
			window:setGlobalLabel(data.total, data.total / duration)
		end
	else
		if O.settings.showFooterOverall == true then
			if window.selectedCombat then
				local data = O.battles[1]:getPreparedPlayerData(window.settings.sort, window.showEnemies)
				local duration = max(O.battles[1].duration, 1)
				window:timerUpdate(duration, "Overall: ")
				window:setGlobalLabel(data.total, data.total / duration)
			else
				window:timerUpdate(0, "Overall: ")
				window:setGlobalLabel(0, 0)
			end
		elseif O.settings.showFooterCurrent == true then
			if window.selectedCombat then
				local data = O.battles[#O.battles]:getPreparedPlayerData(window.settings.sort, window.showEnemies)
				local duration = max(O.battles[#O.battles].duration, 1)
				window:timerUpdate(duration, "Current: ")
				window:setGlobalLabel(data.total, data.total / duration)
			else
				window:timerUpdate(0, "Current: ")
				window:setGlobalLabel(0, 0)
			end
		end
	end
end
function Modes.battle:update(window)
	if not window.selectedCombat then return end

	local rows = {}
	local limit = min(#O.battles, window.settings.rows)

	local nbattles = 1

	--overall total
	--if O.settings.showHistoryTotal == true then
	--local overall = O.overallCombat

	--local duration = max(O.battles[1].duration, 1)
	--local overall = O.battles[1]:getPreparedPlayerData(window.settings.sort, window.showEnemies)
	local duration = max(O.overallCombat.duration, 1)
	local overall = O.overallCombat:getPreparedPlayerData(window.settings.sort, window.showEnemies)
	local orow = {}
	local ostat = overall.total / duration

	O:UpdateInterfaceMenuAlerts( InCombat, self:getPinReportText(window), #O.battles - 1 )

	orow.rightLabel = NumberFormat(floor(ostat))
	orow.value = ostat
	orow.leftClick = function()
  		window:setMode("combat", O.battles[1])
	end
	orow.leftLabel = "T:  " .. FormatSeconds(duration) .. " Total"
	orow.color = O.settings.totalColor
	table.insert( rows, 1, orow )
	--end

	for i = limit, 2, -1 do
		local row = {}
		local combat = O.battles[i + window.scrollOffset]
		
		if not O.settings.showOnlyBoss or (combat.hasBoss or combat == O.CurrentCombat) then
			nbattles = nbattles + 1
			local stat = combat:getPreparedPlayerData(window.settings.sort, window.showEnemies).total / max(combat.duration, 1)
			local hostile = combat:getHostile()

			row.rightLabel = NumberFormat(floor(stat))
			row.value = stat
			row.leftClick = function()
  				window:setMode("combat", O.battles[i + window.scrollOffset])
		   	end
			row.leftLabel = i - 1 + window.scrollOffset .. ":  " .. FormatSeconds(combat.duration) .. " " .. hostile
			rows[nbattles] = row
		end
	end

	if #O.battles > 0 and window.selectedCombat == O.battles[1] then
		if O.settings.showFooterCurrent == true then
			local data = O.CurrentCombat:getPreparedPlayerData(window.settings.sort, window.showEnemies)
			local duration = max(O.CurrentCombat.duration, 1)
			window:timerUpdate(duration, "Current: ")
			window:setGlobalLabel(data.total, data.total / duration)
		elseif O.settings.showFooterOverall == true then
			local data = O.overallCombat:getPreparedPlayerData(window.settings.sort, window.showEnemies)
			local duration = max(O.overallCombat.duration, 1)
			window:timerUpdate(duration, "Overall: ")
			window:setGlobalLabel(data.total, data.total / duration)			
		end
	elseif #O.battles > 0 and window.selectedCombat ~= O.battles[#O.battles] then
		if O.settings.showFooterOverall == true then
			local data = O.overallCombat:getPreparedPlayerData(window.settings.sort, window.showEnemies)
			local duration = max(O.overallCombat.duration, 1)
			window:timerUpdate(duration, "Overall: ")
			window:setGlobalLabel(data.total, data.total / duration)
		elseif O.settings.showFooterCurrent == true then
			local data = O.CurrentCombat:getPreparedPlayerData(window.settings.sort, window.showEnemies)
			local duration = max(O.CurrentCombat.duration, 1)
			window:timerUpdate(duration, "Current: ")
			window:setGlobalLabel(data.total, data.total / duration)
		end
	else
		if O.settings.showFooterOverall == true then
			if window.selectedCombat then
				local data = O.overallCombat:getPreparedPlayerData(window.settings.sort, window.showEnemies)
				local duration = max(O.overallCombat.duration, 1)
				window:timerUpdate(duration, "Overall: ")
				window:setGlobalLabel(data.total, data.total / duration)
			else
				window:timerUpdate(0, "Overall: ")
				window:setGlobalLabel(0, 0)
			end
		elseif O.settings.showFooterCurrent == true then
			if window.selectedCombat then
				local data = O.CurrentCombat:getPreparedPlayerData(window.settings.sort, window.showEnemies)
				local duration = max(O.CurrentCombat.duration, 1)
				window:timerUpdate(duration, "Current: ")
				window:setGlobalLabel(data.total, data.total / duration)
			else
				window:timerUpdate(0, "Current: ")
				window:setGlobalLabel(0, 0)
			end
		end
	end

	return rows, #O.battles, O:GetMaxValueCombat(window.settings.sort, window.showEnemies)
end

function Modes.battle:getPinReportText(window)
	if not window.selectedCombat then return end

	local data = window.selectedCombat:getPreparedPlayerData(window.settings.sort, window.showEnemies)
	local duration = max(window.selectedCombat.duration, 1)
	local text = {}
	tinsert(text, ("Battle Report"))
	tinsert(text, ("Target: %s ~ (%s)"):format(window.selectedCombat:getHostile():sub(0, 16), FormatSeconds(window.selectedCombat.duration)))
	tinsert(text, ("--------------------"))
	tinsert(text, ("Total %s: %d"):format( window.settings.sort, data.total ))
	for i = 1, min( data.count, 5 ) do
		local player = data.players[i]
		if not player.ref.detail.isPet then
			tinsert(text, ("%d: %s - %d (%d) "):format( i, player.ref.detail.name:sub(0, 16), player.value, player.value / duration))
		end
	end

	return table.concat( text, "\n" )
end

function Modes.battle:getReportText(window)
	if not window.selectedCombat then return end

	local data = window.selectedCombat:getPreparedPlayerData(window.settings.sort, window.showEnemies)
	local duration = max(window.selectedCombat.duration, 1)
	local text = {}
	tinsert(text, ("Odyssey Battle Report"))
	tinsert(text, ("Target: %s ~ (%s)"):format(window.selectedCombat:getHostile():sub(0, 16), FormatSeconds(window.selectedCombat.duration)))
	tinsert(text, ("--------------------"))
	tinsert(text, ("Total %s: %d"):format( window.settings.sort, data.total ))
	for i = 1, min( data.count, O.settings.report_lines ) do
		local player = data.players[i]
		if not player.ref.detail.isPet then
			tinsert(text, ("%d: %s - %d (%d) "):format( i, player.ref.detail.name:sub(0, 16), player.value, player.value / duration))
		end
	end

	return text
end

function Modes.combat:mouse4Click(window)
	for i, combat in ipairs(O.battles) do
		if combat == window.selectedCombat and i > 1 then
			window:setMode("battle", O.battles[i - 1])
			return
		end
	end
end

function Modes.combat:mouse5Click(window)
	for i, combat in ipairs(O.battles) do
		if combat == window.selectedCombat and i < #O.battles then
			window:setMode("battle", O.battles[i + 1])
			return
		end
	end
end

Modes.interactions = Mode:new("interactions")
function Modes.interactions:init(window, player)
	if player then
		window.selectedPlayer = player
	end
	if window.selectedPlayer then
		window:setTitle(L["%s: Interactions: %s"]:format(window.selectedPlayer.detail.name, L[window.settings.sort]))
	end
end

function Modes.interactions:update(window)
	if not window.selectedPlayer then
		if window.selectedPlayerDetail then
			self:findOwner(window)
		end
		if not window.selectedPlayer then
			return
		end
	end

	local data = window.selectedPlayer:getInteractions(window.settings.sort)
	local duration = max(window.selectedCombat.duration, 1)
	local rows = {}
	local limit = min(data.count, window.settings.rows)
	for i = 1, limit do
		local row = {}
		local interaction = data.interactions[i + window.scrollOffset]

		row.leftLabel = (O.settings.showRankNumber and i + window.scrollOffset .. ". " or "") .. interaction.name
		row.rightLabel = BuildFormat(NumberFormat(interaction.value), interaction.value / duration, interaction.value / data.total * 100)
		row.color = GetClassColor(interaction.ref.detail)
		row.value = interaction.value
		row.leftClick = function()
			window:setMode("interactionAbilities", interaction.ref, window.selectedPlayer)
		end

		tinsert(rows, row)
	end

	--window:setGlobalLabel(data.total, data.total / duration)

	return rows, data.count, data.max
end

function Modes.interactions:rightClick(window)
	window:setMode("combat")
end

function Modes.interactions:findOwner(window)
	local player = window.selectedCombat.players[window.selectedPlayerDetail]
	if player then
		window.selectedPlayer = player
		window:setTitle(L["%s's Interactions: %s"]:format(player.detail.name, L[window.settings.sort]))
	end
end

Modes.interactionAbilities = Mode:new("interactionAbilities")
function Modes.interactionAbilities:init(window, player, parent)
	if player then
		window.selectedPlayer = player
	end
	if window.selectedPlayer then
		if parent then
			window.selectedPlayerParent = parent
		end
		window:setTitle(("%s: %s: %s"):format(window.selectedPlayerParent.detail.name:sub(0, 6), window.selectedPlayer.detail.name:sub(0, 6), L[window.settings.sort]))
	end
end

function Modes.interactionAbilities:update(window)
	if not window.selectedPlayer then
		if window.selectedPlayerDetail then
			self:findOwner(window)
		end
		if not window.selectedPlayer then
			return
		end
	end

	local data = window.selectedPlayer:getInteractionAbilityData()
	local duration = max(window.selectedCombat.duration, 1)
	local rows = {}
	local limit = min(data.count, window.settings.rows)
	for i = 1, limit do

		-- total bar
		if i == 1 and window.scrollOffset == 0 then
			tinsert(rows, {
				leftLabel = "      " .. L["Total"],
				rightLabel = NumberFormat(data.total),
				value = data.max,
				leftClick = function()
					window:setMode("interactionAbility", window.selectedPlayer:createFakeAbility())
				end
			})
		end

		local row = {}
		local index = window.scrollOffset > 0 and i + window.scrollOffset - 1 or i + window.scrollOffset
		local ability = data.abilities[index]

		row.icon = ability.ref.detail.icon
		row.leftLabel = "      " .. ability.ref.name

		if RelativeTo > 1 then
			row.rightLabel = BuildFormat(NumberFormat(ability.value), ability.value / duration, ability.value / max(data.abilities[RelativeTo - 1].value, 1) * 100)
		else
			row.rightLabel = BuildFormat(NumberFormat(ability.value), ability.value / duration, ability.value / max(window.selectedPlayer[window.settings.sort], 1) * 100)
		end

		row.color = O.settings.abilityTypeColors[ability.ref.type]
		row.value = ability.value
		row.leftClick = function()
			window:setMode("interactionAbility", ability.ref)
		end

		tinsert(rows, row)
	end

	--window:setGlobalLabel(data.total, data.total / duration)

	return rows, data.count + 1, data.max
end

function Modes.interactionAbilities:rightClick(window)
	window:setMode("interactions", window.selectedPlayerParent)
end

function Modes.interactionAbilities:findOwner(window)
	local player = window.selectedCombat.players[window.selectedPlayerDetail]
	if player then
		window.selectedPlayer = player
		window:setTitle(("%s: %s"):format(player.detail.name, L[window.settings.sort]))
	end
end

function Modes.interactionAbilities:onSortmodeChange(window, oldSortmode)
	if window.settings.sort ~= oldSortmode then
		window:setMode("interactions", window.selectedPlayerParent)
		return false
	end
	return true
end

Modes.interactionAbility = Mode:new("interactionAbility")
function Modes.interactionAbility:init(window, ability)
	if ability then
		window.selectedAbility = ability
	end
	if window.selectedAbility then
		window:setTitle(("%s: %s"):format(window.selectedPlayerParent.detail.name, window.selectedAbility.name))
	end
end
function Modes.interactionAbility:update(window)
	if not window.selectedAbility or not window.selectedPlayer then
		self:findOwner(window)
		if not window.selectedAbility then
			return
		end
	end

	local data = window.selectedAbility:getPreparedAbilityStatData(window.settings.sort)

	local rows = {}
	local limit = min(#data, window.settings.rows)
	for i = 1, limit do
		local row = {}
		local stat = data[i + window.scrollOffset]

		row.leftLabel = stat.name
		row.rightLabel = stat.value
		row.color = {0.7, 0.7, 0.7}

		rows[i] = row
	end

	window:setGlobalLabel(window.selectedPlayerParent[window.settings.sort], window.selectedPlayerParent[window.settings.sort] / max(window.selectedCombat.duration, 1))

	return rows, #data, 1
end
function Modes.interactionAbility:rightClick(window)
	window:setMode("interactionAbilities")
end
function Modes.interactionAbility:findOwner(window)
	if not window.selectedPlayer then
		Modes.abilities:findOwner(window)
	end
	if window.selectedPlayer then
		window:setMode("abilities", window.selectedPlayer)
	end
end
function Modes.interactionAbility:onSortmodeChange(window, oldSortmode)
	if window.settings.sort ~= oldSortmode then
		window:setMode("interactions", window.selectedPlayerParent)
		return false
	end
	return true
end

Modes.abilities = Mode:new("abilities")
function Modes.abilities:init(window, player)
	if player then
		window.selectedPlayer = player
	end
	if window.selectedPlayer then
		window:setTitle(("%s: %s"):format(window.selectedPlayer.detail.name, L[window.settings.sort]))
	end
end
function Modes.abilities:update(window)
	if not window.selectedPlayer then
		if window.selectedPlayerDetail then
			self:findOwner(window)
		end
		if not window.selectedPlayer then
			return
		end
	end

	local data = window.selectedPlayer:getPreparedAbilityData(window.settings.sort)
	local duration = max(window.selectedCombat.duration, 1)
	local rows = {}
	local limit = min(data.count, window.settings.rows)
	for i = 1, limit do

		-- total bar
		if i == 1 and self.scrollOffset == 0 then
			tinsert(rows, {
				leftLabel = "      " .. L["Total"],
				rightLabel = NumberFormat(data.total),
				value = data.max,
				leftClick = function()
					window:setMode("ability", window.selectedPlayer:createFakeAbility())
				end
			})
		end

		local row = {}
		local index = window.scrollOffset > 0 and i + window.scrollOffset - 1 or i + window.scrollOffset
		local ability = data.abilities[index]

		row.icon = ability.ref.detail.icon
		row.leftLabel = "      " .. ability.ref.name

		if RelativeTo > 1 then
			row.rightLabel = BuildFormat(NumberFormat(ability.value), ability.value / duration, ability.value / max(data.abilities[RelativeTo - 1].value, 1) * 100)
		else
			row.rightLabel = BuildFormat(NumberFormat(ability.value), ability.value / duration, ability.value / max(window.selectedPlayer[window.settings.sort], 1) * 100)
		end

		row.color = O.settings.abilityTypeColors[ability.ref.type]
		row.value = ability.value
		row.leftClick = function()
			window:setMode("ability", ability.ref)
		end
		
		tinsert(rows, row)
	end

	--window:setGlobalLabel(data.total, data.total / duration)

	return rows, data.count + 1, data.max
end
function Modes.abilities:rightClick(window)
	window:setMode("combat")
end
function Modes.abilities:findOwner(window)
	local player = window.selectedCombat.players[window.selectedPlayerDetail]
	if player then
		window.selectedPlayer = player
		window:setTitle(("%s: %s"):format(player.detail.name, L[window.settings.sort]))
	end
end


Modes.ability = Mode:new("ability")
function Modes.ability:init(window, ability)
	if ability then
		window.selectedAbility = ability
	end
	if window.selectedAbility then
		window:setTitle(("%s: %s"):format(window.selectedPlayer.detail.name, window.selectedAbility.name))
	end
end
function Modes.ability:update(window)
	if not window.selectedAbility or not window.selectedPlayer then
		self:findOwner(window)
		if not window.selectedAbility then
			return
		end
	end

	local ability = window.selectedPlayer:getAbility(window.selectedAbility, window.settings.sort)
	local total
	local data
	if ability then
		data = ability:getPreparedAbilityStatData()
		total = ability.total
	else
		data = window.selectedAbility:getPreparedAbilityStatData(window.settings.sort)
		total = window.selectedPlayer[window.settings.sort]
	end

	local rows = {}
	local limit = min(#data, window.settings.rows)
	for i = 1, limit do
		local row = {}
		local stat = data[i + window.scrollOffset]
		
		row.leftLabel = stat.name
		row.rightLabel = stat.value
		row.color = {0.7, 0.7, 0.7}
		
		rows[i] = row
	end

	--window:setGlobalLabel(total, total / max(window.selectedCombat.duration, 1))

	return rows, #data, 1
end
function Modes.ability:rightClick(window)
	window:setMode("abilities", window.selectedPlayer)
end
function Modes.ability:findOwner(window)
	if not window.selectedPlayer then
		Modes.abilities:findOwner(window)
	end
	if window.selectedPlayer then
		window:setMode("abilities")
	end
end

----------------------------

function O.UI.NewSortmode()
	local sortmode = Sortmode:new()
	tinsert(Sortmodes, sortmode)
	return sortmode
end

function O.UI.Update()
	for i, window in ipairs(Windows) do
		window:update()
	end
end

function O.UI.TimerUpdate(duration)
	local combat = O.battles[1]
	for i, window in ipairs(Windows) do
		if window.selectedCombat == combat then
			if O.settings.showFooterCurrent then
				window:timerUpdate(O.battles[#O.battles].duration, "Current: ")
			end
		else
			if O.settings.showFooterOverall then
				window:timerUpdate(O.battles[1].duration, "Overall: ")
			end
		end
	end
end

function O.UI.NewCombat()
	InCombat = true

		for i, window in ipairs(Windows) do
			-- update to current combat if last combat was selected
			if window.selectedCombat == O.battles[#O.battles - 1] or window.selectedCombat == nil then
				window.selectedCombat = O.battles[#O.battles]

				if window.selectedPlayer then
					window.selectedPlayerDetail = window.selectedPlayer.detail
					window.selectedPlayer = nil
				end
				if window.selectedAbility then
					window.selectedAbilityDetail = window.selectedAbility.detail
					window.selectedAbility = nil
				end

				window:clearRows()
				window.selectedMode:init(window)
			end

			window.frames.buttons.combatStart:Show(false)
			window.frames.buttons.combatEnd:Show(true)
		end

	O:UpdateInterfaceMenuAlerts( InCombat, nil, #O.battles - 1 )

end

function O.UI.EndCombat()
	InCombat = false

	for i, window in ipairs(Windows) do
		window.frames.buttons.combatStart:Show(true)
		window.frames.buttons.combatEnd:Show(false)

		if window.selectedCombat == O.battles[1] then
			if O.settings.showFooterCurrent then
				window:timerUpdate(O.battles[#O.battles].duration, "Current: ")
			end
		else
			if O.settings.showFooterOverall then
				window:timerUpdate(O.battles[1].duration, "Overall: ")
			end
		end
	end

	O:UpdateInterfaceMenuAlerts( InCombat, nil, #O.battles - 1 )

end

function O.UI.Default()
	RemoveOtherWindows()

	Windows[1]:setMode("combat")
	Windows[1]:resize()
end

function O.UI.Reset()
	for i, window in ipairs(Windows) do
		window.selectedCombat = nil
		window:setMode("combat")
		window.frames.timerLabel:SetText("00:00")
		window.frames.globalStatLabel:SetText("0")
		window:clearRows()
	end

	O:UpdateInterfaceMenuAlerts( false, "", nil )

end

function O.UI.Init()
	for i, settings in ipairs(O.settings.windows) do
		tinsert(Windows, Window:new(settings))
	end
end

function O.UI.NewWindow()
	local settings = O.GetDefaultWindowSettings()
	tinsert(O.settings.windows, settings)
	tinsert(Windows, Window:new(settings))
end

function O.UI.Visible(visible)
	for i, window in ipairs(Windows) do
		window.frames.base:Show(visible)
	end
end

function O.UI.ShowResizer(state)
	for i, window in ipairs(Windows) do
		window:showResizer(state)
	end
end

function O.UI.ShowScrollbar(state)
	for i, window in ipairs(Windows) do
		window:showScrollbar(state)
	end
end